import ast
import sys
import pytest
from python_minifier import unparse
from python_minifier.ast_compare import compare_ast

def test_pep():
    if sys.version_info < (3, 8):
        pytest.skip('No Assignment expressions in python < 3.8')

    source = '''
if a := True:
    print(a)
if self._is_special and (ans := self._check_nans(context=context)):
    return ans    
results = [(x, y, x/y) for x in input_data if (y := f(x)) > 0]
stuff = [[y := f(x), x/y] for x in range(5)]    
'''

    expected_ast = ast.parse(source)
    actual_ast = unparse(expected_ast)
    compare_ast(expected_ast, ast.parse(actual_ast))



"""
FString unparsing

This whole module feels like a hack.
Mostly because FStrings feel like a hack.

"""

import ast
import copy

from python_minifier import UnstableMinification
from python_minifier.ast_compare import CompareError
from python_minifier.ast_compare import compare_ast
from python_minifier.expression_printer import ExpressionPrinter
from python_minifier.ministring import MiniString
from python_minifier.rename.util import is_str


class FString(object):
    """
    An F-string in the expression part of another f-string
    """

    def __init__(self, node, allowed_quotes):
        assert isinstance(node, ast.JoinedStr)

        self.node = node
        self.allowed_quotes = allowed_quotes

    def is_correct_ast(self, code):
        try:
            c = ast.parse(code, 'FString candidate', mode='eval')
            compare_ast(self.node, c.body)
            return True
        except Exception as e:
            return False

    def candidates(self):
        actual_candidates = []

        for quote in self.allowed_quotes:
            candidates = ['']
            nested_allowed = copy.copy(self.allowed_quotes)
            nested_allowed.remove(quote)
            for v in self.node.values:
                if is_str(v):
                    try:
                        candidates = [x + self.str_for(v.s, quote) for x in candidates]
                    except Exception as e:
                        continue
                elif isinstance(v, ast.FormattedValue):
                    try:
                        candidates = [
                            x + y for x in candidates for y in FormattedValue(v, nested_allowed).get_candidates()
                        ]
                    except Exception as e:
                        continue
                else:
                    raise RuntimeError('Unexpected JoinedStr value')

                actual_candidates += ['f' + quote + x + quote for x in candidates]

        actual_candidates = filter(self.is_correct_ast, actual_candidates)
        return actual_candidates

    def str_for(self, s, quote):
        return s.replace('{', '{{').replace('}', '}}')


class OuterFString(FString):
    """
    The outermost f-string

    Whereas the FString object assumes backslashes are disallowed, this
    OuterFString is free to use backslashes in the Str parts
    """

    def __init__(self, node):
        assert isinstance(node, ast.JoinedStr)
        super(OuterFString, self).__init__(node, ['"', "'", '"""', "'''"])

    def __str__(self):
        if len(self.node.values) == 0:
            return 'f' + min(self.allowed_quotes, key=len) * 2

        candidates = list(self.candidates())

        for candidate in candidates:

            try:
                minified_f_string = ast.parse(candidate, 'python_minifier.f_string output', mode='eval').body
            except SyntaxError as syntax_error:
                raise UnstableMinification(syntax_error, '', candidate)

            try:
                compare_ast(self.node, minified_f_string)
            except CompareError as compare_error:
                raise UnstableMinification(compare_error, '', candidate)

        if not candidates:
            raise ValueError('Unable to create representation for f-string')

        return min(candidates, key=len)

    def str_for(self, s, quote):
        return str(MiniString(s, quote)).replace('{', '{{').replace('}', '}}')


class FormattedValue(ExpressionPrinter):
    """
    An F-String Expression Part
    """

    def __init__(self, node, allowed_quotes):
        super(FormattedValue, self).__init__()

        assert isinstance(node, ast.FormattedValue)
        self.node = node
        self.allowed_quotes = allowed_quotes
        self.candidates = ['']

    def get_candidates(self):

        self.code = '{'

        if self.is_curly(self.node.value):
            self.code += ' '

        self._expression(self.node.value)

        if self.node.conversion == 115:
            self.code += '!s'
        elif self.node.conversion == 114:
            self.code += '!r'
        elif self.node.conversion == 97:
            self.code += '!a'

        if self.node.format_spec is not None:
            self.code += ':'
            self._append(FormatSpec(self.node.format_spec, self.allowed_quotes).candidates())

        self.code += '}'

        self._finalize()
        return self.candidates

    def is_curly(self, node):
        if isinstance(node, (ast.SetComp, ast.DictComp, ast.Set, ast.Dict)):
            return True

        if isinstance(node, (ast.Expr, ast.Attribute, ast.Subscript)):
            return self.is_curly(node.value)

        if isinstance(node, (ast.Compare, ast.BinOp)):
            return self.is_curly(node.left)

        if isinstance(node, ast.Call):
            return self.is_curly(node.func)

        if isinstance(node, ast.BoolOp):
            return self.is_curly(node.values[0])

        if isinstance(node, ast.IfExp):
            return self.is_curly(node.body)

        return False

    def visit_Str(self, node):
        self.code += str(Str(node.s, self.allowed_quotes))

    def visit_Bytes(self, node):
        self.code += str(Bytes(node.s, self.allowed_quotes))

    def visit_JoinedStr(self, node):
        assert isinstance(node, ast.JoinedStr)
        self.token_break()
        self._append(FString(node, allowed_quotes=self.allowed_quotes).candidates())

    def _finalize(self):
        self.candidates = [x + self.code for x in self.candidates]
        self.code = ''

    def _append(self, candidates):
        self._finalize()
        self.candidates = [x + y for x in self.candidates for y in candidates]


class Str(object):
    """
    A Str node inside an f-string expression

    May use any of the allowed quotes, no backslashes!

    """

    def __init__(self, s, allowed_quotes):
        self._s = s
        self.allowed_quotes = allowed_quotes
        self.current_quote = None

    def _can_quote(self, c):
        if self.current_quote is None:
            return False

        if (c == '\n' or c == '\r') and len(self.current_quote) == 1:
            return False

        if c == self.current_quote[0]:
            return False

        return True

    def _get_quote(self, c):
        for quote in self.allowed_quotes:
            if c == '\n' or c == '\r':
                if len(quote) == 3:
                    return quote
            elif c != quote:
                return quote

        raise ValueError('Couldn\'t find a quote')

    def _literals(self):
        l = ''
        for c in self._s:
            if not self._can_quote(c):
                if l:
                    l += self.current_quote
                    yield l
                    l = ''

                self.current_quote = self._get_quote(c)

            if l == '':
                l += self.current_quote
            l += c

        if l:
            l += self.current_quote
            yield l

    def __str__(self):
        if self._s == '':
            return str(min(self.allowed_quotes, key=len)) * 2

        if '\0' in self._s or '\\' in self._s:
            raise ValueError('Impossible to represent a %r character in f-string expression part')

        if '\n' in self._s or '\r' in self._s:
            if '"""' not in self.allowed_quotes and "'''" not in self.allowed_quotes:
                raise ValueError(
                    'Impossible to represent newline character in f-string expression part without a long quote'
                )

        candidates = []
        for start_quote in self.allowed_quotes:
            self.current_quote = start_quote
            s = ''
            for l in self._literals():
                if s and s[-1] == l[0]:
                    s += ' '
                s += l

            if eval(s) == self._s:
                candidates.append(s)

        if candidates:
            return min(candidates, key=len)
        else:
            raise ValueError('Unable to string')


class FormatSpec(object):
    """
    A FormattedValue format spec

    The AST looks like another f-string. This time there are no quotes.

    """

    def __init__(self, node, allowed_quotes):
        assert isinstance(node, ast.JoinedStr)

        self.node = node
        self.allowed_quotes = allowed_quotes

    def candidates(self):

        candidates = ['']
        for v in self.node.values:
            if is_str(v):
                candidates = [x + self.str_for(v.s) for x in candidates]
            elif isinstance(v, ast.FormattedValue):
                candidates = [
                    x + y for x in candidates for y in FormattedValue(v, self.allowed_quotes).get_candidates()
                ]
            else:
                raise RuntimeError('Unexpected JoinedStr value')

        return candidates

    def str_for(self, s):
        return s.replace('{', '{{').replace('}', '}}')


class Bytes(object):
    """
    A Bytes node inside an f-string expression

    May use any of the allowed quotes, no backslashes!

    """

    def __init__(self, b, allowed_quotes):
        self._b = b
        self.allowed_quotes = allowed_quotes
        self.current_quote = None

    def _can_quote(self, c):
        if self.current_quote is None:
            return False

        if (c == ord(b'\n') or c == ord(b'\r')) and len(self.current_quote) == 1:
            return False

        if chr(c) == self.current_quote[0]:
            return False

        return True

    def _get_quote(self, c):
        for quote in self.allowed_quotes:
            if c == ord(b'\n') or c == ord(b'\r'):
                if len(quote) == 3:
                    return quote
            elif chr(c) != quote:
                return quote

        raise ValueError('Couldn\'t find a quote')

    def _literals(self):
        l = ''
        for b in self._b:
            if not self._can_quote(b):
                if l:
                    l += self.current_quote
                    yield l
                    l = ''

                self.current_quote = self._get_quote(b)

            if l == '':
                l = 'b' + self.current_quote
            l += chr(b)

        if l:
            l += self.current_quote
            yield l

    def __str__(self):
        if self._b == b'':
            return 'b' + str(min(self.allowed_quotes, key=len)) * 2

        if b'\0' in self._b or b'\\' in self._b:
            raise ValueError('Impossible to represent a %r character in f-string expression part')

        if b'\n' in self._b or b'\r' in self._b:
            if '"""' not in self.allowed_quotes and "'''" not in self.allowed_quotes:
                raise ValueError(
                    'Impossible to represent newline character in f-string expression part without a long quote'
                )

        candidates = []
        for start_quote in self.allowed_quotes:
            self.current_quote = start_quote
            s = ''
            for l in self._literals():
                if s and s[-1] == l[0]:
                    s += ' '
                s += l

            assert eval(s) == self._b
            candidates.append(s)

        return min(candidates, key=len)



import ast
import sys


def create_is_namespace():

    namespace_nodes = (ast.FunctionDef, ast.Lambda, ast.ClassDef, ast.Module, ast.GeneratorExp)

    if sys.version_info >= (2, 7):
        namespace_nodes += (ast.SetComp, ast.DictComp)

    if sys.version_info >= (3, 0):
        namespace_nodes += (ast.ListComp,)

    if sys.version_info >= (3, 5):
        namespace_nodes += (ast.AsyncFunctionDef,)

    return lambda node: isinstance(node, namespace_nodes)


is_namespace = create_is_namespace()


def get_global_namespace(node):
    """
    Return the global namespace for a node

    :rtype: :class:`ast.Module`

    """

    if node.namespace is node:
        return node

    return get_global_namespace(node.namespace)


def get_nonlocal_namespace(node):
    """
    Return the nonlocal namespace for a node

    The nonlocal namespace is the closest parent function scope's namespace.
    """

    if isinstance(node.namespace, ast.ClassDef):
        return get_nonlocal_namespace(node.namespace)

    return node.namespace


def arg_rename_in_place(node):
    """
    Can this argument node by safely renamed

    'self', 'cls', 'args', and 'kwargs' are not commonly referenced by the caller, so
    can be safely renamed. Comprehension arguments are not accessible from outside, so
    can be renamed.

    If the argument is positional-only, it can be safely renamed

    Other arguments may be referenced by the caller as keyword arguments, so should not be
    renamed in place. The name assigner may still decide to bind the argument to a new name
    inside the function namespace.

    :param node: The argument node
    :rtype node: :class:`ast.arg`
    :rtype: bool

    """

    func = node.namespace

    if isinstance(func, ast.comprehension):
        return True

    if isinstance(func.namespace, ast.ClassDef) and not isinstance(func, ast.Lambda):
        if len(func.args.args) > 0 and node is func.args.args[0]:
            if len(func.decorator_list) == 0:
                # rename 'self'
                return True
            elif (
                len(func.decorator_list) == 1
                and isinstance(func.decorator_list[0], ast.Name)
                and func.decorator_list[0].id == 'classmethod'
            ):
                # rename 'cls'
                return True

    if func.args.vararg is node or func.args.kwarg is node:
        # starargs
        return True

    if hasattr(func.args, 'posonlyargs') and node in func.args.posonlyargs:
        return True

    return False


def insert(suite, new_node):
    """
    Insert a node into a suite

    Inserts new_node as early as possible in the suite, but after docstrings and `import __future__` statements.

    :param suite: The existing suite to insert the node into
    :param new_node: The node to insert
    :return: :class:`collections.Iterable[Node]`

    """

    inserted = False
    for node in suite:

        if not inserted:
            if (isinstance(node, ast.ImportFrom) and node.module == '__future__') or (
                isinstance(node, ast.Expr) and is_str(node.value)
            ):
                pass
            else:
                yield new_node
                inserted = True

        yield node

    if not inserted:
        yield new_node


def allow_rename_locals(node, rename_locals, preserve_locals=None):

    if preserve_locals is None:
        preserve_locals = []

    if not isinstance(node, ast.Module) and is_namespace(node):
        for binding in node.bindings:
            if rename_locals is False:
                binding.disallow_rename()
            elif binding.name in preserve_locals:
                binding.disallow_rename()

    for child in ast.iter_child_nodes(node):
        allow_rename_locals(child, rename_locals, preserve_locals)


def find__all__(module):

    names = []

    def is_assign_all_node(node):
        if isinstance(node, ast.Assign):
            for name in node.targets:
                if isinstance(name, ast.Name) and name.id == '__all__':
                    return True

        elif isinstance(node, ast.AugAssign) or (hasattr(ast, 'AnnAssign') and isinstance(node, ast.AnnAssign)):
            if isinstance(node.target, ast.Name) and node.target.id == '__all__':
                return True

        return False

    for node in ast.iter_child_nodes(module):
        if not is_assign_all_node(node):
            continue

        if not isinstance(node.value, ast.List):
            continue

        for el in node.value.elts:
            if is_str(el):
                names.append(el.s)

    return names


def allow_rename_globals(module, rename_globals=False, preserve_globals=None):

    if preserve_globals is None:
        preserve_globals = []

    preserve_globals.extend(find__all__(module))

    for binding in module.bindings:
        if rename_globals is False or binding.name in preserve_globals:
            binding.disallow_rename()


def is_str(node):
    if isinstance(node, ast.Str):
        return True

    if hasattr(ast, 'Constant') and isinstance(node, ast.Constant):
        if isinstance(node.value, str):
            return True

    return False


try:
    import builtins
except ImportError:
    import __builtin__ as builtins



import ast

from python_minifier import unparse
from python_minifier.ast_compare import compare_ast

def test_slice():
    """AST for slices was changed in 3.9"""

    source = '''
x[name]
x[1:2]
x[1:2, 3]
x[()]
x[1:2, 2:2]
x[a, ..., b:c]
x[a, ..., b]
x[(a, b)]
x[a:b,]
'''

    expected_ast = ast.parse(source)
    actual_ast = unparse(expected_ast)
    compare_ast(expected_ast, ast.parse(actual_ast))



import ast

from python_minifier.rename.mapper import add_parent


class NodeVisitor(object):
    def visit(self, node):
        """Visit a node."""
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def visit_Constant(self, node):
        if node.value in [None, True, False]:
            method = 'visit_NameConstant'
        elif isinstance(node.value, (int, float, complex)):
            method = 'visit_Num'
        elif isinstance(node.value, str):
            method = 'visit_Str'
        elif isinstance(node.value, bytes):
            method = 'visit_Bytes'
        elif node.value == Ellipsis:
            method = 'visit_Ellipsis'
        else:
            raise RuntimeError('Unknown Constant value %r' % type(node.value))

        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def is_node(self, node, types):
        """
        Is a node one of the specified node types

        A node type may be an actual ast class, or a string naming one.
        types is a single node type or an iterable of many.

        If a node_type specified a specific Constant type (Str, Bytes, Num etc),
        returns true for Constant nodes of the correct type.

        :type node: ast.AST
        :param types:
        :rtype: bool

        """

        if not isinstance(types, tuple):
            types = (types,)

        actual_types = []
        for node_type in types:
            if isinstance(node_type, str):
                node_type = getattr(ast, node_type, None)
                if node_type is not None:
                    actual_types.append(node_type)
            else:
                actual_types.append(node_type)

        if isinstance(node, tuple(actual_types)):
            return True

        if hasattr(ast, 'Constant') and isinstance(node, ast.Constant):
            if node.value in [None, True, False]:
                return ast.NameConstant in types
            elif isinstance(node.value, (int, float, complex)):
                return ast.Num in types
            elif isinstance(node.value, str):
                return ast.Str in types
            elif isinstance(node.value, bytes):
                return ast.Bytes in types
            elif node.value == Ellipsis:
                return ast.Ellipsis in types
            else:
                raise RuntimeError('Unknown Constant value %r' % type(node.value))

        return False


class SuiteTransformer(NodeVisitor):
    """
    Transform suites of instructions
    """

    def __call__(self, node):
        return self.visit(node)

    def visit_ClassDef(self, node):
        node.bases = [self.visit(b) for b in node.bases]

        node.body = self.suite(node.body, parent=node)
        node.decorator_list = [self.visit(d) for d in node.decorator_list]

        if hasattr(node, 'starargs') and node.starargs is not None:
            node.starargs = self.visit(node.starargs)

        if hasattr(node, 'kwargs') and node.kwargs is not None:
            node.kwargs = self.visit(node.kwargs)

        if hasattr(node, 'keywords'):
            node.keywords = [self.visit(kw) for kw in node.keywords]

        return node

    def visit_FunctionDef(self, node):
        node.args = self.visit(node.args)
        node.body = self.suite(node.body, parent=node)
        node.decorator_list = [self.visit(d) for d in node.decorator_list]

        if hasattr(node, 'returns') and node.returns is not None:
            node.returns = self.visit(node.returns)

        return node

    def visit_AsyncFunctionDef(self, node):
        return self.visit_FunctionDef(node)

    def visit_For(self, node):
        node.target = self.visit(node.target)
        node.iter = self.visit(node.iter)

        node.body = self.suite(node.body, parent=node)

        if node.orelse:
            node.orelse = self.suite(node.orelse, parent=node)

        return node

    def visit_AsyncFor(self, node):
        return self.visit_For(node)

    def visit_If(self, node):
        node.test = self.visit(node.test)

        node.body = self.suite(node.body, parent=node)

        if node.orelse:
            node.orelse = self.suite(node.orelse, parent=node)

        return node

    def visit_Try(self, node):
        node.body = self.suite(node.body, parent=node)

        node.handlers = [self.visit(h) for h in node.handlers]

        if node.orelse:
            node.orelse = self.suite(node.orelse, parent=node)

        if node.finalbody:
            node.finalbody = self.suite(node.finalbody, parent=node)

        return node

    def visit_While(self, node):
        node.test = self.visit(node.test)

        node.body = self.suite(node.body, parent=node)

        if node.orelse:
            node.orelse = self.suite(node.orelse, parent=node)

        return node

    def visit_With(self, node):

        if hasattr(node, 'items'):
            node.items = [self.visit(i) for i in node.items]
        else:
            if node.context_expr:
                node.context_expr = self.visit(node.context_expr)
            if node.optional_vars:
                node.optional_vars = self.visit(node.optional_vars)

        node.body = self.suite(node.body, parent=node)
        return node

    def visit_AsyncWith(self, node):
        return self.visit_With(node)

    def visit_Module(self, node):
        node.body = self.suite(node.body, parent=node)
        return node

    def suite(self, node_list, parent):
        return [self.visit(node) for node in node_list]

    def generic_visit(self, node):
        for field, old_value in ast.iter_fields(node):
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, ast.AST):
                        value = self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, ast.AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node

    def add_child(self, child, parent, namespace=None):
        def nearest_function_namespace(node):
            """
            Return the namespace node for the nearest function scope.

            This could be itself.

            :param node: The node to get the function namespace of
            :type node: ast.Node
            :rtype: ast.Node

            """

            if isinstance(node, (ast.FunctionDef, ast.Module)):
                return node
            if hasattr(ast, 'AsyncFunctionDef') and isinstance(node, ast.AsyncFunctionDef):
                return node
            return nearest_function_namespace(node.parent)

        if namespace is None:
            namespace = nearest_function_namespace(parent)

        add_parent(child, parent=parent, namespace=namespace)
        return child



import ast

from python_minifier import add_namespace
from python_minifier.transforms.combine_imports import CombineImports
from python_minifier.ast_compare import compare_ast

def combine_imports(module):
    add_namespace(module)
    CombineImports()(module)
    return module

def test_import():
    source = '''import builtins
import collections'''
    expected = 'import builtins, collections'


    expected_ast = ast.parse(expected)
    actual_ast = combine_imports(ast.parse(source))
    compare_ast(expected_ast, actual_ast)

def test_import_as():
    source = '''import builtins
import collections as c
import functools as f
import datetime

pass'''
    expected = '''import builtins, collections as c, functools as f, datetime
pass'''


    expected_ast = ast.parse(expected)
    actual_ast = combine_imports(ast.parse(source))
    compare_ast(expected_ast, actual_ast)


def test_import_from():
    source = '''from builtins import dir
from builtins import help
import collections
from collections import abc'''
    expected = '''from builtins import dir, help 
import collections
from collections import abc'''

    expected_ast = ast.parse(expected)
    actual_ast = combine_imports(ast.parse(source))
    compare_ast(expected_ast, actual_ast)

def test_import_in_function():
    source = '''def test():
    import collection as c
    import builtins
    
    return None
'''
    expected = '''def test():
    import collection as c, builtins
    return None
'''

    expected_ast = ast.parse(expected)
    actual_ast = combine_imports(ast.parse(source))
    compare_ast(expected_ast, actual_ast)


def test_import_star():
    source = '''
from breakfast import hashbrown
from breakfast import *
from breakfast import sausage
from breakfast import bacon
'''
    expected = '''
from breakfast import hashbrown
from breakfast import *
from breakfast import sausage, bacon
'''

    expected_ast = ast.parse(expected)
    actual_ast = combine_imports(ast.parse(source))
    compare_ast(expected_ast, actual_ast)



import itertools

from python_minifier.rename.name_generator import name_filter

def test_name_generator():

    ng = name_filter()

    names = set()
    for name in itertools.islice(ng, 10000):
        assert len(name) <= 3
        names.add(name)

    assert len(names) == 10000

    # Check no keywords returned
    assert 'or' not in names

    # Check no builtins returned
    assert 'id' not in names
    assert 'abs' not in names



import ast

from python_minifier.rename.util import arg_rename_in_place, insert


class Binding(object):
    """
    Represents the binding of a name

    :param name: A name for this binding
    :type name: str or None
    :param bool allow_rename: If this binding may be renamed
    :param int rename_cost: The cost of renaming this binding in bytes, NOT including the difference in name lengths

    """

    def __init__(self, name=None, allow_rename=True, rename_cost=0):
        self._references = []

        self._allow_rename = allow_rename
        self._rename_cost = rename_cost

        self._name = name
        self._reserved = None

    def __repr__(self):
        return self.__class__.__name__ + '()'

    @property
    def name(self):
        """
        The name for this binding

        This may be changed using the rename() method.
        If this binding doesn't currently have a name, this returns None.

        :rtype: str or None

        """

        return self._name

    @property
    def allow_rename(self):
        """
        Is it allowed to rename this binding

        :rtype: bool

        """

        return self._allow_rename

    def disallow_rename(self):
        """
        Prevent this binding from being renamed
        """

        self._allow_rename = False

    @property
    def reserved(self):
        """
        A reserved name for this binding

        This may be a name which this binding reserves in it's reservation scope,
        regardless of if it is renamed.

        :rtype: str or None

        """

        return self._reserved

    @property
    def references(self):
        """
        The ast Nodes that reference this binding

        :rtype: list[ast.AST]

        """

        return self._references

    @property
    def name_references(self):
        """
        The number of times the name is used
        """
        return len(self._references)

    def add_reference(self, node, allow_rename=True, reserved=None, rename_cost=0):
        """
        Add a new reference to this binding

        :param node: The node that references this binding
        :type node: :class:`ast.AST`
        :param bool allow_rename: If this binding may be renamed
        :param str reserved: A name used by the node, even if the binding is renamed.
        :param int rename_cost: Additional cost of renaming the reference, in bytes

        """

        self.references.append(node)

        if allow_rename is False:
            self.disallow_rename()

        if reserved is not None:
            self._reserved = reserved

        self._rename_cost += rename_cost

    def should_rename(self, new_name):
        """
        Is it space efficient to rename this binding

        :param str new_name: The candidate name
        :rtype: bool

        """

        raise NotImplementedError()

    def rename(self, new_name):
        """
        Rename this binding and all nodes that reference it

        :param str new_name: The new name to use

        """

        raise NotImplementedError()


class NameBinding(Binding):
    """
    Represents the binding of a defined name

    A NameBinding will be attached to the local namespace that defines it.

    :param str name: The original bound name
    :param bool allow_rename: If this binding may be renamed
    :param int rename_cost: The cost of renaming this binding in bytes

    """

    def __init__(self, name, *args, **kwargs):
        super(NameBinding, self).__init__(name, *args, **kwargs)

        if name.startswith('__') and name.endswith('__'):
            # System defined name
            self.disallow_rename()

    def __repr__(self):
        return self.__class__.__name__ + '(name=%r) <references=%r>' % (self._name, len(self._references))

    def should_rename(self, new_name):
        """
        Is it space efficient to rename this binding

        :param str new_name: The candidate name
        :rtype: bool

        """

        current_cost = len(self.references) * len(self._name)
        rename_cost = (len(self.references) * len(new_name)) + self._rename_cost

        return rename_cost <= current_cost

    def disallow_rename(self):
        """
        Prevent this binding from being renamed
        """

        super(NameBinding, self).disallow_rename()
        self._reserved = self._name

    def rename(self, new_name):
        """
        Rename this binding and all nodes that reference it

        :param str new_name: The new name to use

        """

        for node in self.references:

            if isinstance(node, ast.Name):

                if isinstance(node.ctx, (ast.Load, ast.Store, ast.Del)):
                    node.id = new_name
                else:
                    # Python 2 Param context

                    if arg_rename_in_place(node):
                        node.id = new_name

                    else:
                        node.namespace.body = list(
                            insert(
                                node.namespace.body,
                                ast.Assign(
                                    targets=[ast.Name(id=new_name, ctx=ast.Store())],
                                    value=ast.Name(id=node.id, ctx=ast.Load()),
                                ),
                            )
                        )

            elif isinstance(node, ast.FunctionDef) or (
                hasattr(ast, 'AsyncFunctionDef') and isinstance(node, ast.AsyncFunctionDef)
            ):
                node.name = new_name
            elif isinstance(node, ast.ClassDef):
                node.name = new_name
            elif isinstance(node, ast.alias):
                if new_name == node.name:
                    node.asname = None
                else:
                    node.asname = new_name
            elif hasattr(ast, 'arg') and isinstance(node, ast.arg):

                if arg_rename_in_place(node):
                    node.arg = new_name

                else:
                    node.namespace.body = list(
                        insert(
                            node.namespace.body,
                            ast.Assign(
                                targets=[ast.Name(id=new_name, ctx=ast.Store())],
                                value=ast.Name(id=node.arg, ctx=ast.Load()),
                            ),
                        )
                    )

            elif isinstance(node, ast.ExceptHandler):
                node.name = new_name
            elif isinstance(node, ast.Global) or (hasattr(ast, 'Nonlocal') and isinstance(node, ast.Nonlocal)):
                node.names = [new_name if n == self._name else n for n in node.names]
            elif isinstance(node, ast.arguments):

                rename_vararg = (node.vararg == self._name) and not getattr(node, 'vararg_renamed', False)
                rename_kwarg = (node.kwarg == self._name) and not getattr(node, 'kwarg_renamed', False)

                if rename_vararg:
                    node.vararg = new_name
                    node.vararg_renamed = True
                if rename_kwarg:
                    node.kwarg = new_name
                    node.kwarg_renamed = True

        self._name = new_name


class BuiltinBinding(NameBinding):
    """
    Represents the usage of a builtin

    :param str name: The name of the builtin
    :param namespace: The module the builtin is used in
    :type namespace: :class:`ast.Module`

    """

    def __init__(self, name, namespace, *args, **kwargs):
        super(BuiltinBinding, self).__init__(name, *args, **kwargs)
        self.namespace = namespace

        # These builtins actually act like keywords, so should not be changed
        if name == 'super':
            # If we replace 'super' with another name the compiler will neglect to create the
            # __class__ implicit closure reference, breaking the zero argument super() call.
            self.disallow_rename()
        elif name == 'object':
            # Classes must inherit from object to become a new-style class in python2
            self.disallow_rename()

    def rename(self, new_name):
        builtin = self._name
        super(BuiltinBinding, self).rename(new_name)
        self.namespace.body = list(
            insert(
                self.namespace.body,
                ast.Assign(
                    targets=[ast.Name(id=new_name, ctx=ast.Store())], value=ast.Name(id=builtin, ctx=ast.Load())
                ),
            )
        )



import ast

from python_minifier.rename.binding import BuiltinBinding, NameBinding
from python_minifier.rename.util import get_global_namespace, get_nonlocal_namespace, builtins


def get_binding(name, namespace):
    if name in namespace.global_names and not isinstance(namespace, ast.Module):
        return get_binding(name, get_global_namespace(namespace))
    elif name in namespace.nonlocal_names and not isinstance(namespace, ast.Module):
        return get_binding(name, get_nonlocal_namespace(namespace))

    for binding in namespace.bindings:
        if binding.name == name:
            return binding

    if not isinstance(namespace, ast.Module):
        return get_binding(name, get_nonlocal_namespace(namespace))

    else:
        # This is unresolved at global scope - is it a builtin?
        if name in dir(builtins):
            if name in ['exec', 'eval', 'locals', 'globals', 'vars']:
                namespace.tainted = True

            binding = BuiltinBinding(name, namespace, rename_cost=len(name) + 1)
            namespace.bindings.append(binding)
            return binding

        else:
            binding = NameBinding(name)
            binding.disallow_rename()
            namespace.bindings.append(binding)
            return binding


def resolve_names(node):
    """
    Resolve unbound names to a NameBinding

    :param node: The module to resolve names in
    :type node: :class:`ast.Module`

    """

    if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
        get_binding(node.id, node.namespace).add_reference(node)
    elif hasattr(ast, 'Exec') and isinstance(node, ast.Exec):
        get_global_namespace(node).tainted = True

    for child in ast.iter_child_nodes(node):
        resolve_names(child)



import ast
from python_minifier import add_namespace, bind_names, resolve_names
from python_minifier.transforms.remove_pass import RemovePass
from python_minifier.ast_compare import compare_ast

def remove_literals(source):
    module = ast.parse(source, 'remove_literals')

    add_namespace(module)
    bind_names(module)
    resolve_names(module)
    return RemovePass()(module)

def test_remove_pass_empty_module():
    source = 'pass'
    expected = ''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_pass_module():
    source = '''import collections
pass
a = 1
pass'''
    expected = '''import collections
a=1'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_if_empty():
    source = '''if True:
    pass'''
    expected = '''if True:
    0'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_if_line():
    source = '''if True: pass'''
    expected = '''if True: 0'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_suite():
    source = '''if True: 
    pass
    a=1 
    pass 
    return None'''
    expected = '''if True:
    a=1 
    return None'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_from_class():
    source = '''class A:
    pass
    a = 1
    pass
    def b():
        pass
        return 1
        pass
'''
    expected = '''class A:
    a=1
    def b():
        return 1
'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_from_class_empty():
    source = '''class A:
    pass
'''
    expected = 'class A:0'

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_from_class_func_empty():
    source = '''class A:
    def b():
        pass
'''
    expected = '''class A:
    def b(): 0'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)



import ast

from .expression_printer import ExpressionPrinter


class ModulePrinter(ExpressionPrinter):
    """
    Builds the smallest possible exact representation of an ast
    """

    def __init__(self, indent_char='\t'):
        super(ModulePrinter, self).__init__()
        self.indent_char = indent_char

    def __call__(self, module):
        """
        Generate the source code for an AST

        :param module: The Module to generate code for
        :type module: ast.Module
        :rtype: str

        """

        assert isinstance(module, ast.Module)

        self.visit_Module(module)
        self.code = self.code.rstrip('\n' + self.indent_char + ';')
        return self.code

    def newline(self):
        """
        Ensure there is a newline at the end of the output
        """
        if self.code == '':
            return

        self.code = self.code.rstrip('\n' + self.indent_char + ';')
        self.code += '\n'
        self.code += self.indent_char * self.indent

    # region Simple Statements

    def visit_Exec(self, node):
        assert isinstance(node, ast.Exec)

        self.token_break()

        self.code += 'exec'
        self._expression(node.body)

        if node.globals:
            self.token_break()
            self.code += 'in'
            self._expression(node.globals)

        if node.locals:
            self.code += ','
            self._expression(node.locals)

        self.end_statement()

    def visit_Expr(self, node):
        assert isinstance(node, ast.Expr)

        if isinstance(node.value, ast.Yield):
            self._yield_expr(node.value)
        elif hasattr(ast, 'YieldFrom') and isinstance(node.value, ast.YieldFrom):
            self._yield_expr(node.value)
        else:
            self._testlist(node.value)

        self.end_statement()

    def visit_Assert(self, node):
        assert isinstance(node, ast.Assert)

        self.token_break()

        self.code += 'assert'
        self._expression(node.test)

        if node.msg:
            self.code += ','
            self._expression(node.msg)

        self.end_statement()

    def visit_Assign(self, node):
        assert isinstance(node, ast.Assign)

        for target_node in node.targets:
            self._testlist(target_node)
            self.code += '='

        # Yield nodes that are the sole node on the right hand side of an assignment do not need parens
        if isinstance(node.value, ast.Yield):
            self._yield_expr(node.value)
        elif hasattr(ast, 'YieldFrom') and isinstance(node.value, ast.YieldFrom):
            self._yield_expr(node.value)
        else:
            self._testlist(node.value)

        self.end_statement()

    def visit_AugAssign(self, node):
        assert isinstance(node, ast.AugAssign)

        self._testlist(node.target)
        self.visit(node.op)
        self.code += '='

        # Yield nodes that are the sole node on the right hand side of an assignment do not need parens
        if isinstance(node.value, ast.Yield):
            self._yield_expr(node.value)
        elif hasattr(ast, 'YieldFrom') and isinstance(node.value, ast.YieldFrom):
            self._yield_expr(node.value)
        else:
            self._testlist(node.value)

        self.end_statement()

    def visit_AnnAssign(self, node):
        assert isinstance(node, ast.AnnAssign)

        if node.simple:
            self.visit(node.target)
        else:
            self.code += '('
            self._expression(node.target)
            self.code += ')'

        if node.annotation:
            self.code += ':'
            self._expression(node.annotation)

        if node.value:
            self.code += '='

            self._expression(node.value)

        self.end_statement()

    def visit_Pass(self, node):
        assert isinstance(node, ast.Pass)

        self.token_break()
        self.code += 'pass'
        self.end_statement()

    def visit_Delete(self, node):
        assert isinstance(node, ast.Delete)

        self.code += 'del'
        self._exprlist(node.targets)
        self.end_statement()

    def visit_Return(self, node):
        assert isinstance(node, ast.Return)

        self.token_break()
        self.code += 'return'
        if node.value is not None:
            self._testlist(node.value)
        self.end_statement()

    def visit_Print(self, node):
        assert isinstance(node, ast.Print)

        self.code += 'print'

        first = True

        if node.dest:
            self.code += '>>'
            self._expression(node.dest)
            first = False

        for v in node.values:
            if first:
                first = False
            else:
                self.code += ','

            self._expression(v)

        if not node.nl:
            self.code += ','

        self.end_statement()

    def visit_Yield(self, node):
        assert isinstance(node, ast.Yield)

        self._yield_expr(node)
        self.end_statement()

    def visit_YieldFrom(self, node):
        assert isinstance(node, ast.YieldFrom)

        self._yield_expr(node)
        self.end_statement()

    def visit_Raise(self, node):
        assert isinstance(node, ast.Raise)

        self.code += 'raise'

        if hasattr(node, 'type'):
            # Python2 raise node

            if node.type:
                self.code += ' '
                self._expression(node.type)
            if node.inst:
                self.code += ','
                self._expression(node.inst)
            if node.tback:
                self.code += ','
                self._expression(node.tback)

        else:
            # Python3

            if node.exc:
                self.code += ' '
                self._expression(node.exc)

            if node.cause:
                self.code += ' from '
                self._expression(node.cause)

        self.end_statement()

    def visit_Break(self, node):
        assert isinstance(node, ast.Break)

        self.token_break()
        self.code += 'break'
        self.end_statement()

    def visit_Continue(self, node):
        assert isinstance(node, ast.Continue)

        self.token_break()
        self.code += 'continue'
        self.end_statement()

    def visit_Import(self, node):
        assert isinstance(node, ast.Import)

        self.code += 'import '

        first = True
        for n in node.names:
            if first:
                first = False
            else:
                self.code += ','

            self.visit_alias(n)

        self.end_statement()

    def visit_ImportFrom(self, node):
        assert isinstance(node, ast.ImportFrom)

        if node.module is None:
            self.code += 'from ' + ('.' * node.level) + ' '
        else:
            self.code += 'from ' + ('.' * node.level)
            self.code += node.module

        self.code += ' import '
        first = True
        for n in node.names:
            if first:
                first = False
            else:
                self.code += ','

            if node.module == '__future__' and n.name == 'unicode_literals':
                self.unicode_literals = True

            self.visit_alias(n)

        self.end_statement()

    def visit_alias(self, node):
        assert isinstance(node, ast.alias)

        self.code += node.name

        if node.asname:
            self.code += ' as ' + node.asname

    def visit_Global(self, node):
        assert isinstance(node, ast.Global)

        self.code += 'global ' + ','.join(node.names)
        self.end_statement()

    def visit_Nonlocal(self, node):
        assert isinstance(node, ast.Nonlocal)

        self.code += 'nonlocal ' + ','.join(node.names)
        self.end_statement()

    # endregion

    # region Compound Statements

    def visit_If(self, node, el=False):
        assert isinstance(node, ast.If)

        self.newline()

        if el:
            self.code += 'el'
        self.code += 'if'
        self._expression(node.test)
        self.code += ':'

        self._suite(node.body)

        if node.orelse:
            if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If):
                # elif
                self.visit_If(node.orelse[0], el=True)
                self.newline()
            else:
                # an else block
                self.code += 'else:'
                self._suite(node.orelse)

    def visit_For(self, node, is_async=False):
        assert isinstance(node, ast.For) or (hasattr(ast, 'AsyncFor') and isinstance(node, ast.AsyncFor))

        self.newline()

        if is_async:
            self.code += 'async '

        self.code += 'for '
        self._exprlist([node.target])
        self.code += ' in '
        self._expression(node.iter)
        self.code += ':'

        self._suite(node.body)

        if node.orelse:
            self.newline()
            self.code += 'else:'
            self._suite(node.orelse)

    def visit_While(self, node):
        assert isinstance(node, ast.While)

        self.newline()
        self.code += 'while '
        self._expression(node.test)
        self.code += ':'
        self._suite(node.body)

        if node.orelse:
            self.code += 'else:'
            self._suite(node.orelse)

    def visit_Try(self, node):
        assert isinstance(node, ast.Try)

        self.newline()
        self.code += 'try:'
        self._suite(node.body)

        [self.visit_ExceptHandler(n) for n in node.handlers]

        if node.orelse:
            self.code += 'else:'
            self._suite(node.orelse)

        if node.finalbody:
            self.code += 'finally:'
            self._suite(node.finalbody)

    def visit_TryFinally(self, node):
        assert isinstance(node, ast.TryFinally)

        if len(node.body) == 1 and isinstance(node.body[0], ast.TryExcept):
            self.visit_TryExcept(node.body[0])
        else:
            self.newline()
            self.code += 'try:'
            self._suite(node.body)

        if node.finalbody:
            self.code += 'finally:'
            self._suite(node.finalbody)

    def visit_TryExcept(self, node):
        assert isinstance(node, ast.TryExcept)

        self.newline()
        self.code += 'try:'
        self._suite(node.body)

        [self.visit_ExceptHandler(n) for n in node.handlers]

        if node.orelse:
            self.code += 'else:'
            self._suite(node.orelse)

    def visit_ExceptHandler(self, node):
        assert isinstance(node, ast.ExceptHandler)

        self.code += 'except'
        if node.type is not None:
            self.code += ' '
            self._expression(node.type)

        if node.name is not None:
            self.token_break()
            self.code += 'as'

            if isinstance(node.name, str):
                self.code += ' ' + node.name
            else:
                self._expression(node.name)

        self.code += ':'

        self._suite(node.body)

    def visit_With(self, node, is_async=False):
        assert isinstance(node, ast.With) or (hasattr(ast, 'AsyncWith') and isinstance(node, ast.AsyncWith))

        self.newline()

        if is_async:
            self.code += 'async '

        self.code += 'with'

        first = True
        if hasattr(node, 'items'):

            for item in node.items:
                if first:
                    first = False
                else:
                    self.code += ','

                if self.precedence(item.context_expr) != 0 and self.precedence(item.context_expr) <= self.precedence(
                    node
                ):
                    self.code += '('
                    self.visit_withitem(item)
                    self.code += ')'
                else:
                    self.visit_withitem(item)
        else:
            self.visit_withitem(node)

        self.code += ':'
        self._suite(node.body)

    def visit_withitem(self, node):
        assert (hasattr(ast, 'withitem') and isinstance(node, ast.withitem)) or isinstance(node, ast.With)

        self._expression(node.context_expr)

        if node.optional_vars is not None:
            self.token_break()
            self.code += 'as'
            self._expression(node.optional_vars)

    def visit_FunctionDef(self, node, is_async=False):
        assert isinstance(node, ast.FunctionDef) or (
            hasattr(ast, 'AsyncFunctionDef') and isinstance(node, ast.AsyncFunctionDef)
        )

        self.newline()

        for d in node.decorator_list:
            self.code += '@'
            self._expression(d)
            self.newline()

        if is_async:
            self.code += 'async '

        self.code += 'def ' + node.name + '('
        self.visit_arguments(node.args)
        self.code += ')'

        if hasattr(node, 'returns') and node.returns is not None:
            self.code += '->'
            self._expression(node.returns)
            self.code += ':'
        else:
            self.code += ':'

        if hasattr(node, 'docstring') and node.docstring is not None:
            self._suite([ast.Expr(value=ast.Str(s=node.docstring))] + node.body)
        else:
            self._suite(node.body)

    def visit_ClassDef(self, node):
        assert isinstance(node, ast.ClassDef)

        self.newline()

        for d in node.decorator_list:
            self.code += '@'
            self._expression(d)
            self.newline()

        first = True
        self.code += 'class ' + node.name

        for b in node.bases:
            if first:
                self.code += '('
                first = False
            else:
                self.code += ','
            self._expression(b)

        if hasattr(node, 'starargs') and node.starargs is not None:
            if first:
                self.code += '('
                first = False
            else:
                self.code += ','

            self.code += '*'
            self._expression(node.starargs)

        if hasattr(node, 'keywords'):
            for kw in node.keywords:
                if first:
                    self.code += '('
                    first = False
                else:
                    self.code += ','
                self.visit_keyword(kw)

        if hasattr(node, 'kwargs') and node.kwargs is not None:
            if first:
                self.code += '('
                first = False
            else:
                self.code += ','

            self.code += '**'
            self.visit(node.kwargs)

        if not first:
            self.code += ')'

        self.code += ':'

        if hasattr(node, 'docstring') and node.docstring is not None:
            self._suite([ast.Expr(value=ast.Str(s=node.docstring))] + node.body)
        else:
            self._suite(node.body)

    # endregion

    # region async and await

    def visit_AsyncFunctionDef(self, node):
        assert isinstance(node, ast.AsyncFunctionDef)
        self.visit_FunctionDef(node, is_async=True)

    def visit_AsyncFor(self, node):
        assert isinstance(node, ast.AsyncFor)
        self.visit_For(node, is_async=True)

    def visit_AsyncWith(self, node):
        assert isinstance(node, ast.AsyncWith)
        self.visit_With(node, is_async=True)

    # endregion

    def visit_Module(self, node):
        if hasattr(node, 'docstring') and node.docstring is not None:
            # Python 3.6 added a docstring field! Really useful for every use case except this one...
            # Put the docstring back into the body
            self._suite_body([ast.Expr(value=ast.Str(s=node.docstring))] + node.body)
        else:
            self._suite_body(node.body)

    def _suite(self, node_list):

        compound_statements = [
            'For',
            'While',
            'Try',
            'If',
            'With',
            'ClassDef',
            'TryFinally',
            'TryExcept',
            'FunctionDef',
            'AsyncFunctionDef',
            'AsyncFor',
            'AsyncWith',
        ]

        if [node for node in node_list if node.__class__.__name__ in compound_statements]:
            self.enter_block()
            self._suite_body(node_list)
            self.leave_block()
        else:
            self.indent += 1
            self._suite_body(node_list)
            self.indent -= 1
            self.newline()

    def _suite_body(self, node_list):

        statements = {
            'Assign': self.visit_Assign,
            'AnnAssign': self.visit_AnnAssign,
            'AugAssign': self.visit_AugAssign,
            'Expr': self.visit_Expr,
            'Delete': self.visit_Delete,
            'Pass': self.visit_Pass,
            'Import': self.visit_Import,
            'ImportFrom': self.visit_ImportFrom,
            'Global': self.visit_Global,
            'Nonlocal': self.visit_Nonlocal,
            'Assert': self.visit_Assert,
            'Break': self.visit_Break,
            'Continue': self.visit_Continue,
            'Return': self.visit_Return,
            'Raise': self.visit_Raise,
            'Yield': self.visit_Yield,
            'YieldFrom': self.visit_YieldFrom,
            'For': self.visit_For,
            'While': self.visit_While,
            'Try': self.visit_Try,
            'If': self.visit_If,
            'With': self.visit_With,
            'ClassDef': self.visit_ClassDef,
            'FunctionDef': self.visit_FunctionDef,
            'AsyncFunctionDef': self.visit_AsyncFunctionDef,
            'AsyncFor': self.visit_AsyncFor,
            'AsyncWith': self.visit_AsyncWith,
            'TryFinally': self.visit_TryFinally,
            'TryExcept': self.visit_TryExcept,
            'Print': self.visit_Print,
            'Exec': self.visit_Exec,
        }

        for node in node_list:
            statements[node.__class__.__name__](node)

    def enter_block(self):
        self.indent += 1
        self.newline()

    def leave_block(self):
        self.indent -= 1
        self.newline()

    def end_statement(self):
        """ End a statement with a newline, or a semi-colon if it saves characters. """

        if self.indent == 0:
            self.newline()
        else:
            if self.code[-1] != ';':
                self.code += ';'



import ast


class CompareError(RuntimeError):
    """
    Raised when an AST compares unequal.
    """

    def __init__(self, lnode, rnode, msg=None):
        self.lnode = lnode
        self.rnode = rnode
        self.msg = msg

    def __repr__(self):
        return 'NodeError(%r, %r)' % (self.lnode, self.rnode)

    def namespace(self, node):
        if hasattr(node, 'namespace'):
            if isinstance(node.namespace, (ast.FunctionDef, ast.ClassDef)) or (
                hasattr(ast, 'AsyncFunctionDef') and isinstance(node.namespace, ast.AsyncFunctionDef)
            ):
                return self.namespace(node.namespace) + '.' + node.namespace.name
            elif isinstance(node.namespace, ast.Module):
                return ''
            else:
                return repr(node.namespace.__class__)

        return None

    def __str__(self):
        error = ''

        if self.msg:
            error += self.msg

        if self.namespace(self.lnode):
            error += ' in namespace ' + self.namespace(self.lnode)

        if self.lnode and hasattr(self.lnode, 'lineno'):
            error += ' at source %i:%i' % (self.lnode.lineno, self.lnode.col_offset)

        return error


def compare_ast(l_ast, r_ast):
    """
    Compare Python Abstract Syntax Trees

    >>> compare_ast(l_ast, r_ast)

    If the AST's are not identical, an exception will be raised.

    """

    def counter():
        i = 0
        while True:
            yield i
            i += 1

    if type(l_ast) != type(r_ast):
        raise CompareError(l_ast, r_ast, msg='Nodes do not match! %r != %r' % (l_ast, r_ast))

    for field in set(l_ast._fields + r_ast._fields):

        if field == 'kind' and isinstance(l_ast, ast.Constant):
            continue

        if isinstance(getattr(l_ast, field, None), list):

            l_list = getattr(l_ast, field, None)
            r_list = getattr(r_ast, field, None)

            if len(l_list) != len(r_list):
                raise CompareError(
                    l_list,
                    r_list,
                    'List does not have the same number of elements! len(%s.%s)=%r, len(%s.%s)=%r'
                    % (type(l_ast), field, len(l_list), type(r_ast), field, len(r_list)),
                )

            for i, l, r in zip(counter(), l_list, r_list):
                if isinstance(l, ast.AST) or isinstance(r, ast.AST):
                    compare_ast(l, r)
                elif l != r:
                    raise CompareError(
                        l_ast,
                        r_ast,
                        'Fields do not match! %s.%s[%i]=%r, %s.%s[%i]=%r'
                        % (type(l_ast), field, i, l, type(r_ast), field, i, r),
                    )

        else:
            l = getattr(l_ast, field, None)
            r = getattr(r_ast, field, None)

            if isinstance(l, ast.AST) or isinstance(r, ast.AST):
                compare_ast(l, r)
            elif l != r:
                raise CompareError(
                    l_ast,
                    r_ast,
                    'Fields do not match! %s.%s=%r, %s.%s=%r' % (type(l_ast), field, l, type(r_ast), field, r),
                )



import ast

from python_minifier.rename.binding import NameBinding
from python_minifier.rename.util import arg_rename_in_place, get_global_namespace, get_nonlocal_namespace, builtins
from python_minifier.transforms.suite_transformer import NodeVisitor


class NameBinder(NodeVisitor):
    """
    Create a NameBinding for each name that is bound

    The NameBinding is added to the bindings dictionary in the namespace node the name is local to.
    """

    def __call__(self, module):
        assert isinstance(module, ast.Module)
        module.tainted = False
        return self.visit(module)

    def get_binding(self, name, namespace):
        if name in namespace.global_names and not isinstance(namespace, ast.Module):
            return self.get_binding(name, get_global_namespace(namespace))
        elif name in namespace.nonlocal_names and not isinstance(namespace, ast.Module):
            return self.get_binding(name, get_nonlocal_namespace(namespace))

        if isinstance(namespace, ast.ClassDef):
            binding = self.get_binding(name, get_nonlocal_namespace(namespace))
            binding.disallow_rename()
            return binding

        for binding in namespace.bindings:
            if binding.name == name:
                break
        else:  # weeee!
            binding = NameBinding(name)
            namespace.bindings.append(binding)

            if name in dir(builtins):
                binding.disallow_rename()

        if name in namespace.nonlocal_names and isinstance(namespace, ast.Module):
            # This is actually a syntax error - but we want the same syntax error after minifying!
            binding.disallow_rename()

        return binding

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store) or isinstance(node.ctx, ast.Del):
            self.get_binding(node.id, node.namespace).add_reference(node)

        if isinstance(node.ctx, ast.Param):
            binding = self.get_binding(node.id, node.namespace)

            if arg_rename_in_place(node):
                binding.add_reference(node)
            else:
                binding.add_reference(
                    node, reserved=node.id, rename_cost=(len(node.id) * 2) + 2
                )  # Assuming that the new name is only a single character

                if isinstance(node.namespace, ast.Lambda):
                    # Lambda function arguments can't be renamed without breaking keyword arguments
                    binding.disallow_rename()

    def visit_ClassDef(self, node):
        self.get_binding(node.name, node.namespace).add_reference(node)
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self.get_binding(node.name, node.namespace).add_reference(node)
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        self.visit_FunctionDef(node)

    def visit_alias(self, node):
        if node.name == '*':
            get_global_namespace(node).tainted = True

        root_module = node.name.split('.')[0]

        if root_module == 'timeit':
            get_global_namespace(node).tainted = True

        if node.asname is not None:
            self.get_binding(node.asname, node.namespace).add_reference(node)
        else:
            # This binds the root module only for a dotted import

            binding = self.get_binding(root_module, node.namespace)
            binding.add_reference(node, rename_cost=len(node.name) + 4)

            if '.' in node.name:
                binding.disallow_rename()

    def visit_arguments(self, node):
        if isinstance(node.vararg, str):
            binding = self.get_binding(node.vararg, node.namespace)
            binding.add_reference(node)

        if isinstance(node.kwarg, str):
            binding = self.get_binding(node.kwarg, node.namespace)
            binding.add_reference(node)

        self.generic_visit(node)

    def visit_arg(self, node):
        binding = self.get_binding(node.arg, node.namespace)

        if arg_rename_in_place(node):
            binding.add_reference(node)
        else:
            binding.add_reference(
                node, reserved=node.arg, rename_cost=(len(node.arg) * 2) + 2
            )  # Assuming that the new name is only a single character

            if isinstance(node.namespace, ast.Lambda):
                # Lambda function arguments can't be renamed without breaking keyword arguments
                binding.disallow_rename()

        self.generic_visit(node)

    def visit_ExceptHandler(self, node):
        if node.name is not None:
            if isinstance(node.name, str):
                # python 3
                self.get_binding(node.name, node.namespace).add_reference(node)
            else:
                # In python 2 the name is a Name node,
                # which will be visited by generic_visit
                pass

        self.generic_visit(node)

    def visit_Global(self, node):
        for name in node.names:
            self.get_binding(name, node.namespace).add_reference(node)

    def visit_Nonlocal(self, node):
        for name in node.names:
            self.get_binding(name, node.namespace).add_reference(node)


def bind_names(module):
    """
    Bind names to their local namespace

    :param module: The module to bind names in
    :type: :class:`ast.Module`

    """

    NameBinder()(module)



import ast

from python_minifier.rename.binding import NameBinding
from python_minifier.rename.name_generator import name_filter
from python_minifier.rename.util import is_namespace


def all_bindings(node):
    """
    All bindings in a module

    :param node: The module to get bindings in
    :type node: :class:`ast.AST`
    :rtype: Iterable[ast.AST, Binding]

    """

    if is_namespace(node):
        for binding in node.bindings:
            yield node, binding

    for child in ast.iter_child_nodes(node):
        for namespace, binding in all_bindings(child):
            yield namespace, binding


def sorted_bindings(module):
    """
    All bindings in a modules sorted by descending number of references

    :param module: The module to get bindings in
    :type module: :class:`ast.AST`
    :rtype: Iterable[ast.AST, Binding]

    """

    def comp(tup):
        namespace, binding = tup
        return len(binding.references)

    return sorted(all_bindings(module), key=comp, reverse=True)


def reservation_scope(namespace, binding):
    """
    Get the namespaces that are in the bindings reservation scope

    Returns the namespace nodes the binding name must be resolvable in

    :param namespace: The local namespace of a binding
    :type namespace: :class:`ast.AST`
    :param binding: The binding to get the reservation scope for
    :type binding: Binding
    :rtype: set[ast.AST]

    """

    namespaces = set([namespace])

    for node in binding.references:
        while node is not namespace:
            namespaces.add(node.namespace)
            node = node.namespace

    return namespaces


def add_assigned(node):
    """
    Add the assigned_names attribute to namespace nodes in a module

    :param node: The module to add the assigned_names attribute to
    :type node: :class:`ast.Module`

    """

    if is_namespace(node):
        node.assigned_names = set()

    for child in ast.iter_child_nodes(node):
        add_assigned(child)


def reserve_name(name, reservation_scope):
    """
    Reserve a name in a reservation scope

    :param str name: The name to reserve
    :param reservation_scope:
    :type reservation_scope: Iterable[:class:`ast.AST`]

    """

    for namespace in reservation_scope:
        namespace.assigned_names.add(name)


class UniqueNameAssigner(object):
    """
    Assign new names to renamed bindings

    Assigns a unique name to every binding
    """

    def __init__(self):
        self.name_generator = name_filter()
        self.names = []

    def available_name(self):
        return next(self.name_generator)

    def __call__(self, module):
        assert isinstance(module, ast.Module)

        for namespace, binding in sorted_bindings(module):
            if binding.allow_rename:
                binding.new_name = self.available_name()

        return module


class NameAssigner(object):
    """
    Assign new names to renamed bindings

    This assigner creates a name 'reservation scope' containing each namespace a binding is referenced in, including
    transitive namespaces. Bindings are then assigned the first available name that has no references in their
    reservation scope. This means names will be reused in sibling namespaces, and shadowed where possible in child
    namespaces.

    Bindings are assigned names in order of most references, with names assigned shortest first.

    """

    def __init__(self, name_generator=None):
        self.name_generator = name_generator if name_generator is not None else name_filter()
        self.names = []

    def iter_names(self):
        for name in self.names:
            yield name

        while True:
            name = next(self.name_generator)
            self.names.append(name)
            yield name

    def available_name(self, reservation_scope, prefix=''):
        """
        Search for the first name that is not in reservation scope
        """

        for name in self.iter_names():
            if self.is_available(prefix + name, reservation_scope):
                return prefix + name

    def is_available(self, name, reservation_scope):
        """
        Is a name unreserved in a reservation scope

        :param str name: the name to check availability of
        :param reservation_scope: The scope to check
        :type reservation_scope: Iterable[:class:`ast.AST`]
        :rtype: bool

        """

        for namespace in reservation_scope:
            if name in namespace.assigned_names:
                return False

        return True

    def __call__(self, module, prefix_globals, reserved_globals=None):
        assert isinstance(module, ast.Module)
        add_assigned(module)

        for namespace, binding in all_bindings(module):
            if binding.reserved is not None:
                scope = reservation_scope(namespace, binding)
                reserve_name(binding.reserved, scope)

        if reserved_globals is not None:
            for name in reserved_globals:
                module.assigned_names.add(name)

        for namespace, binding in sorted_bindings(module):
            scope = reservation_scope(namespace, binding)

            if binding.allow_rename:

                if isinstance(namespace, ast.Module) and prefix_globals:
                    name = self.available_name(scope, prefix='_')
                else:
                    name = self.available_name(scope)

                def should_rename():
                    if binding.should_rename(name):
                        return True

                    # It's no longer efficient to do this rename

                    if isinstance(binding, NameBinding):
                        # Check that the original name is still available

                        if binding.reserved == binding.name:
                            # We already reserved it (this is probably an arg)
                            return False

                        if not self.is_available(binding.name, scope):
                            # The original name has already been assigned to another binding,
                            # so we need to rename this anyway.
                            return True

                    return False

                if should_rename():
                    binding.rename(name)
                else:
                    # Any existing name will become reserved
                    binding.disallow_rename()

            if binding.name is not None:
                reserve_name(binding.name, scope)

        return module


def rename(module, prefix_globals=False, preserved_globals=None):
    NameAssigner()(module, prefix_globals, preserved_globals)



"""
Test for renaming of local names

This assumes the standard NameAssigner and name_generator
"""

import ast
import sys

import pytest

from python_minifier import add_namespace, bind_names, resolve_names, allow_rename_locals, allow_rename_globals, \
    compare_ast, rename, CompareError, unparse


def rename_locals(source):

    # This will raise if the source file can't be parsed
    module = ast.parse(source, 'test_rename_locals')
    add_namespace(module)
    bind_names(module)
    resolve_names(module)

    allow_rename_locals(module, True)
    allow_rename_globals(module, False)

    rename(module)

    return module

def assert_code(expected_ast, actual_ast):
    try:
        compare_ast(expected_ast, actual_ast)
    except CompareError as e:
        print(e)
        print(unparse(actual_ast))
        raise


def test_rename_variable_single_scope():
    source = '''
def a():
    a = 2
'''
    expected = '''
def a():
    A = 2
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_rename_multiple_definitions():
    source = '''
def inner(my_name):
    import collections as my_name
    def my_name(): pass
    my_name = 12
'''
    expected = '''
def inner(my_name):
    A=my_name
    import collections as A
    def A(): pass
    A = 12
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_rename_self_cls_in_place():
    source = '''
class TestClass():
    def mymethod(self): return self
    def mymethod2(qwertyuiop): return qwertyuiop
    def mymethod3(): return 'No Self'
    def mymethod4(self, a): return self, a

    @staticmethod
    def mystatic(cls): return cls
    @staticmethod    
    def mystatic(asdfghjkl): return asdfghjkl
    @staticmethod    
    def mystatic(): return 'No cls'
    
    @classmethod
    def mystatic(cls): return cls
    @classmethod    
    def mystatic(asdfghjkl): return asdfghjkl
    @classmethod    
    def mystatic(): return 'No cls'
    
    @unknown_decorator
    def unknown(self): return self
    @unknown_decorator    
    def unknown(qwertyuiop): return qwertyuiop
    @unknown_decorator        
    def unknown(self, arg): return self, arg
    @unknown_decorator    
    def unknown(): return 'No arg'        
    
'''
    expected = '''
class TestClass():
    def mymethod(A): return A
    def mymethod2(A): return A
    def mymethod3(): return 'No Self'
    def mymethod4(A, a): return A, a

    @staticmethod
    def mystatic(cls): return cls
    @staticmethod    
    def mystatic(asdfghjkl): return asdfghjkl
    @staticmethod    
    def mystatic(): return 'No cls'
    
    @classmethod
    def mystatic(A): return A
    @classmethod    
    def mystatic(A): return A
    @classmethod    
    def mystatic(): return 'No cls'
    
    @unknown_decorator
    def unknown(self): return self
    @unknown_decorator    
    def unknown(qwertyuiop): return qwertyuiop
    @unknown_decorator        
    def unknown(self, arg): return self, arg
    @unknown_decorator    
    def unknown(): return 'No arg'        
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)


def test_rename_arg_kwargs_in_place():
    source = '''
def test(arg, arg2, *args, **kwargs): return args, kwargs
def samename(arg, arg2, *asd, **asd): return asd
    
class TestClass():
    def mymethod(self, arg, *args, **kwargs): return args, kwargs
    
    @classmethod
    def mymethod(cls, arg, *args, **kwargs): return args, kwargs
'''
    expected = '''
def test(arg, arg2, *A, **B): return A, B
def samename(arg, arg2, *A, **A): return A
    
class TestClass():
    def mymethod(C, arg, *A, **B): return A, B
    
    @classmethod
    def mymethod(C, arg, *A, **B): return A, B   
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_no_rename_long_arg():
    source = '''
def f(this_is_my_long_argument_name):
    print(this_is_my_long_argument_name)
'''

    expected_ast = ast.parse(source)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)


def test_rename_long_arg():
    source = '''
def f(this_is_my_long_argument_name):
    print(this_is_my_long_argument_name)
    print(this_is_my_long_argument_name)    
'''
    expected = '''
def f(this_is_my_long_argument_name):
    A = this_is_my_long_argument_name
    print(A)
    print(A)
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_no_rename_single_char_arg():
    source = '''
def f(a):
    print(a)
    print(a)
    print(a)
    print(a) 
    print(a)  
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a) 
    print(a)                     
'''

    expected_ast = ast.parse(source)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_rename_arg():
    source = '''
def f(aa):
    print(aa)
    print(aa)
    print(aa)
    print(aa)
    print(aa)                           
'''
    expected = '''
def f(aa):
    A = aa
    print(A)
    print(A)
    print(A)
    print(A)
    print(A)                   
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_no_rename_lambda_arg():
    source = '''
lambda my_argument: f(my_argument + my_argument + my_argument)                       
'''

    expected_ast = ast.parse(source)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_rename_lambda_stararg():
    source = '''
lambda *args, **kwargs: f(args, kwargs)                       
'''

    expected = '''
lambda *A, **B: f(A, B)                       
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_python3_listcomp_scope():
    if sys.version_info < (3, 0):
        pytest.skip('No list comprehension scope in python < 3.0')

    source = '''
[a for a in mylist]                     
'''

    expected = '''
[A for A in mylist]                     
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_python2_listcomp_scope():
    if sys.version_info >= (3, 0):
        pytest.skip('list comprehension scope in python >= 3.0')

    source = '''
[a for a in mylist]               
def t():
    a = True
    [a for a in mylist] 
'''

    expected = '''
[a for a in mylist]               
def t():
    A = True
    [A for A in mylist]                    
'''

    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)

def test_arg_rename():

    source = '''
def f(*B,**A):pass
'''
    expected = '''
def f(*A,**B):pass
'''
    expected_ast = ast.parse(expected)
    actual_ast = rename_locals(source)
    assert_code(expected_ast, actual_ast)



import ast
import sys

from python_minifier.transforms.suite_transformer import SuiteTransformer


class RemoveObject(SuiteTransformer):
    def __call__(self, node):
        if sys.version_info < (3, 0):
            return node

        return self.visit(node)

    def visit_ClassDef(self, node):
        node.bases = [
            b for b in node.bases if not isinstance(b, ast.Name) or (isinstance(b, ast.Name) and b.id != 'object')
        ]

        node.body = [self.visit(n) for n in node.body]

        return node



import ast
import sys
import pytest
from python_minifier import unparse
from python_minifier.ast_compare import compare_ast

def test_pep():
    if sys.version_info < (3, 9):
        pytest.skip('Decorator expression not allowed in python <3.9')

    source = """
buttons = [QPushButton(f'Button {i}') for i in range(10)]

# Do stuff with the list of buttons...

@buttons[0].clicked.connect
def spam():
    ...

@buttons[1].clicked.connect
def eggs():
    ...

# Do stuff with the list of buttons...
@(f, g)
def a(): pass

@(f, g)
class A:pass

@lambda func: (lambda *p: func(*p).u())
def g(n): pass

@s := lambda func: (lambda *p: func(*p).u())
def g(name): pass

@s
def r(n, t):
    pass

@lambda f: lambda *p: f or f(*p).u()
def g(name): pass

@lambda f: lambda *p: \
        [_ for _ in [ \
            f(*p),
            ] if _][0]
def c(): pass

@lambda f: lambda *p: \
            list(filter(lambda _: _,[
                (a := t()) and False,
                f(*p),
                (b := t()) and False,
            ]))[0]
def c(): pass

"""

    expected_ast = ast.parse(source)
    actual_ast = unparse(expected_ast)
    compare_ast(expected_ast, ast.parse(actual_ast))



import ast
import sys


class ExpressionPrinter(object):
    """
    Builds the smallest possible exact representation of an ast
    """

    def __init__(self):

        self.code = ''
        self.indent = 0
        self.unicode_literals = False

        self.precedences = {
            'Lambda': 2,
            'IfExp': 3,
            'comprehension': 3.5,
            'Or': 4,
            'And': 5,
            'Not': 6,
            'In': 7, 'NotIn': 7, 'Is': 7, 'IsNot': 7, 'Lt': 7, 'LtE': 7, 'Gt': 7, 'GtE': 7, 'NotEq': 7, 'Eq': 7,
            'BitOr': 8,
            'BitXor': 9,
            'BitAnd': 10,
            'LShift': 11, 'RShift': 11,
            'Add': 12, 'Sub': 12,
            'Mult': 13, 'Div': 13, 'FloorDiv': 13, 'Mod': 13, 'MatMult': 13,
            'UAdd': 14, 'USub': 14, 'Invert': 14,
            'Pow': 15,
            'Await': 16,
            'Subscript': 17, 'Call': 17, 'Attribute': 17,
            'Tuple': 18, 'Set': 18, 'List': 18, 'Dict': 18,
        }

    def __call__(self, module):
        """
        Generate the source code for an AST

        :param module: The Module to generate code for
        :type module: ast.Module
        :rtype: str

        """

        self.visit(module)
        return self.code

    def precedence(self, node):
        """
        The precedence of an expression

        Node will usually be an operator or literal.
        Nodes with no precedence value return 0.

        :param node: The AST node to decide precedence for
        :type node: ast.Node
        :rtype: int

        """

        if isinstance(node, (ast.BinOp, ast.UnaryOp, ast.BoolOp)):
            return self.precedences[node.op.__class__.__name__]
        elif isinstance(node, ast.Compare):
            return min(self.precedences[n.__class__.__name__] for n in node.ops)

        # Python2 parses negative ints as an ast.Num with a negative value.
        # Make sure the Num get the precedence of the USub operator in this case.
        if sys.version_info < (3, 0) and isinstance(node, ast.Num):
            if str(node.n)[0] == '-':
                return self.precedences['USub']

        return self.precedences.get(node.__class__.__name__, 0)

    def visit(self, node):
        """
        Visit a node

        Call the correct visit_ method based on the node type.
        Prefer to call the correct method directly if you already know
        the node type.

        :param node: The node to visit
        :type node: ast.Node

        """

        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.visit_Unknown)
        return visitor(node)

    def visit_Unknown(self, node):
        raise RuntimeError('Unknown node %r' % node)

    # region Literals

    def visit_Constant(self, node):
        if node.value in [None, True, False]:
            return self.visit_NameConstant(node)
        elif isinstance(node.value, (int, float, complex)):
            return self.visit_Num(node)
        elif isinstance(node.value, str):
            return self.visit_Str(node)
        elif isinstance(node.value, bytes):
            return self.visit_Bytes(node)
        elif node.value == Ellipsis:
            return self.visit_Ellipsis(node)

        raise RuntimeError('Unknown Constant value %r' % type(node.value))

    def visit_Num(self, node):
        self.token_break()

        v = repr(node.n)

        if v == 'inf':
            self.code += '1e999'
        elif v == '-inf':
            self.code += '-1e999'
        elif v in ['infj', 'inf*j']:
            self.code += '1e999j'
        elif v in ['-infj', '-inf*j']:
            self.code += '-1e999j'

        else:
            if isinstance(node.n, int):
                # Due to the 0x notation, it's unlikely a base-16 literal will be more compact than base-10
                # But for those rare cases....
                h = hex(node.n)
                if len(h) < len(v):
                    v = h

            self.code += v

    def visit_Str(self, node):

        s = repr(node.s)

        if sys.version_info < (3, 0) and self.unicode_literals:
            if s[0] == 'u':
                s = s[1:]
            else:
                s = 'b' + s

        if len(s) > 0 and s[0].isalpha():
            self.token_break()

        self.code += s

    def visit_Bytes(self, node):

        s = repr(node.s)

        if len(s) > 0 and s[0].isalpha():
            self.token_break()

        self.code += s

    def visit_List(self, node):
        self.code += '['
        self._exprlist(node.elts)
        self.code += ']'

    def visit_Tuple(self, node):

        if len(node.elts) == 0:
            self.code += '()'
            return

        self._exprlist(node.elts)

        if len(node.elts) == 1:
            self.code += ','

    def visit_Set(self, node):
        self.code += '{'
        self._exprlist(node.elts)
        self.code += '}'

    def visit_Dict(self, node):
        self.code += '{'

        first = True
        for k, v in zip(node.keys, node.values):
            if not first:
                self.code += ','
            else:
                first = False

            if k is None:
                self.code += '**'
            else:
                self._expression(k)
                self.code += ':'

            self._expression(v)

        self.code += '}'

    def visit_Ellipsis(self, node):
        self.code += '...'

    def visit_NameConstant(self, node):
        self.token_break()
        self.code += repr(node.value)

    # endregion

    # region Variables

    def visit_Name(self, node):
        self.token_break()
        self.code += node.id

    def visit_Starred(self, node):
        self.code += '*'
        self._expression(node.value)

    # endregion

    # region Expressions

    def visit_UnaryOp(self, node):
        self.visit(node.op)

        if sys.version_info < (3, 0) and isinstance(node.op, ast.USub) and isinstance(node.operand, ast.Num):
            # For: -(1), which is parsed as a UnaryOp(USub, Num(1)).
            # Without this special case it would be printed as -1
            # This is fine, but python 2 will then parse it at Num(-1) so the AST wouldn't round-trip.

            self.code += '('
            self.visit_Num(node.operand)
            self.code += ')'
            return

        self._rhs(node.operand, node)

    def visit_UAdd(self, node):
        self.code += '+'

    def visit_USub(self, node):
        self.code += '-'

    def visit_Not(self, node):
        self.token_break()
        self.code += 'not'

    def visit_Invert(self, node):
        self.code += '~'

    def visit_BinOp(self, node):
        self._lhs(node.left, node.op)
        self.visit(node.op)
        self._rhs(node.right, node.op)

    def visit_Add(self, node):
        self.code += '+'

    def visit_Sub(self, node):
        self.code += '-'

    def visit_Mult(self, node):
        self.code += '*'

    def visit_Div(self, node):
        self.code += '/'

    def visit_FloorDiv(self, node):
        self.code += '//'

    def visit_Mod(self, node):
        self.code += '%'

    def visit_Pow(self, node):
        self.code += '**'

    def visit_LShift(self, node):
        self.code += '<<'

    def visit_RShift(self, node):
        self.code += '>>'

    def visit_BitOr(self, node):
        self.code += '|'

    def visit_BitXor(self, node):
        self.code += '^'

    def visit_BitAnd(self, node):
        self.code += '&'

    def visit_MatMult(self, node):
        self.code += '@'

    def visit_BoolOp(self, node):
        first = True

        op_precedence = self.precedence(node.op)

        for v in node.values:
            if first:
                first = False
            else:
                self._expression(node.op)

            value_precedence = self.precedence(v)

            if value_precedence != 0 and (
                (op_precedence > value_precedence)
                or op_precedence == value_precedence
                and self._is_left_associative(node.op)
            ):
                self.code += '('
                self._expression(v)
                self.code += ')'
            else:
                self._expression(v)

    def visit_And(self, node):
        self.token_break()
        self.code += 'and'

    def visit_Or(self, node):
        self.token_break()
        self.code += 'or'

    def visit_Compare(self, node):

        left_precedence = self.precedence(node.left)
        op_precedence = self.precedence(node.ops[0])

        if left_precedence != 0 and ((op_precedence > left_precedence) or (op_precedence == left_precedence)):
            self.code += '('
            self._expression(node.left)
            self.code += ')'
        else:
            self._expression(node.left)

        for op, comparator in zip(node.ops, node.comparators):
            self._expression(op)
            self._rhs(comparator, op)

    def visit_Eq(self, node):
        self.code += '=='

    def visit_NotEq(self, node):
        self.code += '!='

    def visit_Lt(self, node):
        self.code += '<'

    def visit_LtE(self, node):
        self.code += '<='

    def visit_Gt(self, node):
        self.code += '>'

    def visit_GtE(self, node):
        self.code += '>='

    def visit_Is(self, node):
        self.token_break()
        self.code += 'is'

    def visit_IsNot(self, node):
        self.token_break()
        self.code += 'is not'

    def visit_In(self, node):
        self.token_break()
        self.code += 'in'

    def visit_NotIn(self, node):
        self.token_break()
        self.code += 'not in'

    def visit_Call(self, node):

        self._lhs(node.func, node)

        self.code += '('

        first = True
        for arg in node.args:
            if first:
                first = False
            else:
                self.code += ','

            self._expression(arg)

        if node.keywords:
            for kwarg in node.keywords:
                if first:
                    first = False
                else:
                    self.code += ','

                assert isinstance(kwarg, ast.keyword)
                self.visit_keyword(kwarg)

        if hasattr(node, 'starargs') and node.starargs is not None:
            if first:
                first = False
            else:
                self.code += ','

            self.code += '*'
            self._expression(node.starargs)

        if hasattr(node, 'kwargs') and node.kwargs is not None:
            if not first:
                self.code += ','

            self.code += '**'
            self.visit(node.kwargs)

        self.code += ')'

    def visit_keyword(self, node):
        if node.arg is None:
            self.code += '**'
            self._expression(node.value)
        else:
            self.code += node.arg + '='
            self._expression(node.value)

    def visit_IfExp(self, node):

        self._rhs(node.body, node)

        self.token_break()
        self.code += 'if'

        self._rhs(node.test, node)

        self.token_break()
        self.code += 'else'

        self._expression(node.orelse)

    def visit_Attribute(self, node):
        self.token_break()

        value_precedence = self.precedence(node.value)
        attr_precedence = self.precedence(node)

        if (value_precedence != 0 and (attr_precedence > value_precedence)) or isinstance(node.value, ast.Num):
            self.code += '('
            self._expression(node.value)
            self.code += ')'
        else:
            self._expression(node.value)

        self.code += '.' + node.attr

    # endregion

    # region Subscripting

    def visit_Subscript(self, node):

        value_precedence = self.precedence(node.value)
        slice_precedence = 17  # self.precedence(node)

        if value_precedence != 0 and (slice_precedence > value_precedence):
            self.code += '('
            self._expression(node.value)
            self.code += ')'
        else:
            self._expression(node.value)

        self.code += '['

        if isinstance(node.slice, ast.Index):
            self.visit_Index(node.slice)
        elif isinstance(node.slice, ast.Slice):
            self.visit_Slice(node.slice)
        elif isinstance(node.slice, ast.ExtSlice):
            self.visit_ExtSlice(node.slice)
        elif isinstance(node.slice, ast.Ellipsis):
            self.visit_Ellipsis(node)
        elif sys.version_info >= (3, 9) and isinstance(node.slice, ast.Tuple):
            self.visit_Tuple(node.slice)
        elif sys.version_info >= (3, 9):
            self._expression(node.slice)
        else:
            raise AssertionError('Unknown slice type %r' % node.slice)

        self.code += ']'

    def visit_Index(self, node):
        self._expression(node.value)

    def visit_Slice(self, node):
        if node.lower:
            self._expression(node.lower)
        self.code += ':'
        if node.upper:
            self._expression(node.upper)
        if node.step:
            self.code += ':'
            self._expression(node.step)

    def visit_ExtSlice(self, node):
        first = True

        for s in node.dims:
            if not first:
                self.code += ','
            else:
                first = False

            self._expression(s)

        if len(node.dims) == 1:
            self.code += ','

    # endregion

    # region Comprehensions

    def visit_ListComp(self, node):
        self.code += '['
        self._expression(node.elt)
        [self.visit_comprehension(x) for x in node.generators]
        self.code += ']'

    def visit_SetComp(self, node):
        self.code += '{'
        self._expression(node.elt)
        [self.visit_comprehension(x) for x in node.generators]
        self.code += '}'

    def visit_GeneratorExp(self, node):
        self.code += '('
        self._expression(node.elt)
        [self.visit_comprehension(x) for x in node.generators]
        self.code += ')'

    def visit_DictComp(self, node):
        self.code += '{'
        self._expression(node.key)
        self.code += ':'
        self._expression(node.value)
        [self.visit_comprehension(x) for x in node.generators]
        self.code += '}'

    def visit_comprehension(self, node):
        assert isinstance(node, ast.comprehension)

        self.token_break()

        if hasattr(node, 'is_async') and node.is_async:
            self.code += 'async '

        self.code += 'for'
        self._exprlist([node.target])
        self.token_break()
        self.code += 'in'

        self._rhs(node.iter, node)

        if node.ifs:
            for i in node.ifs:
                self.token_break()
                self.code += 'if'
                self._rhs(i, node)

    # endregion

    # region Function and Class definitions

    def visit_Lambda(self, node):

        self.token_break()
        self.code += 'lambda'

        self.visit_arguments(node.args)

        self.code += ':'

        self._expression(node.body)

    def visit_arguments(self, node):
        first = True

        args = getattr(node, 'posonlyargs', []) + node.args

        count_no_defaults = len(args) - len(node.defaults)
        for i, arg in enumerate(args):
            if not first:
                self.code += ','
            else:
                self.token_break()
                first = False

            self._expression(arg)

            if i >= count_no_defaults:
                self.code += '='
                self._expression(node.defaults[i - count_no_defaults])

            if hasattr(node, 'posonlyargs') and node.posonlyargs and i + 1 == len(node.posonlyargs):
                self.code += ',/'

        if node.vararg:
            if not first:
                self.code += ','
            else:
                self.token_break()
                first = False

            self.code += '*'

            if hasattr(node, 'varargannotation'):
                self.code += node.vararg
                if node.varargannotation is not None:
                    self.code += ':'
                    self._expression(node.varargannotation)
            elif isinstance(node.vararg, str):
                self.code += node.vararg
            else:
                self.visit(node.vararg)

        if hasattr(node, 'kwonlyargs') and node.kwonlyargs:

            if not node.vararg:
                if not first:
                    self.code += ','
                else:
                    self.token_break()
                    first = False

                self.code += '*'

            for i, arg in enumerate(node.kwonlyargs):
                self.code += ','
                self.visit_arg(arg)

                if node.kw_defaults[i] is not None:
                    self.code += '='
                    self._expression(node.kw_defaults[i])

        if node.kwarg:
            if not first:
                self.code += ','
            else:
                self.token_break()
                first = False

            self.code += '**'

            if hasattr(node, 'kwargannotation'):
                self.code += node.kwarg
                if node.kwargannotation is not None:
                    self.code += ':'
                    self._expression(node.kwargannotation)
            elif isinstance(node.kwarg, str):
                self.code += node.kwarg
            else:
                self.visit(node.kwarg)

    def visit_arg(self, node):
        if isinstance(node, ast.Name):
            # Python 2 uses Name nodes
            return self.visit_Name(node)

        self.code += node.arg

        if node.annotation:
            self.code += ':'
            self._expression(node.annotation)

    def visit_Repr(self, node):
        self.code += '`'
        self._expression(node.value)
        self.code += '`'

    # endregion

    def visit_Expression(self, node):
        self._expression(node.body)

    def _expression(self, expression):
        if isinstance(expression, ast.Yield) or (hasattr(ast, 'YieldFrom') and isinstance(expression, ast.YieldFrom)):
            self.code += '('
            self._yield_expr(expression)
            self.code += ')'
        elif isinstance(expression, ast.Tuple) and len(expression.elts) > 0:
            self.code += '('
            self.visit_Tuple(expression)
            self.code += ')'
        elif hasattr(ast, 'NamedExpr') and isinstance(expression, ast.NamedExpr):
            self.code += '('
            self.visit_NamedExpr(expression)
            self.code += ')'
        else:
            self.visit(expression)

    def _testlist(self, test):
        if isinstance(test, ast.Yield) or (hasattr(ast, 'YieldFrom') and isinstance(test, ast.YieldFrom)):
            self.code += '('
            self._yield_expr(test)
            self.code += ')'
        elif hasattr(ast, 'NamedExpr') and isinstance(test, ast.NamedExpr):
            self.code += '('
            self.visit_NamedExpr(test)
            self.code += ')'
        else:
            self.visit(test)

    def _exprlist(self, exprlist):
        first = True

        for expr in exprlist:
            if first:
                first = False
            else:
                self.code += ','
            self._expression(expr)

    def _yield_expr(self, yield_node):
        self.token_break()

        if isinstance(yield_node, ast.Yield):
            self.code += 'yield'
        elif isinstance(yield_node, ast.YieldFrom):
            self.code += 'yield from'

        if yield_node.value is not None:
            self._expression(yield_node.value)

    @staticmethod
    def _is_right_associative(operator):
        return isinstance(operator, ast.Pow)

    @staticmethod
    def _is_left_associative(operator):
        return not isinstance(operator, ast.Pow)

    def _lhs(self, left_node, op_node):
        left_precedence = self.precedence(left_node)
        op_precedence = self.precedence(op_node)

        if left_precedence != 0 and (
            (op_precedence > left_precedence)
            or (op_precedence == left_precedence and self._is_right_associative(op_node))
        ):
            self.code += '('
            self._expression(left_node)
            self.code += ')'
        else:
            self._expression(left_node)

    def _rhs(self, right_node, op_node):
        right_precedence = self.precedence(right_node)
        op_precedence = self.precedence(op_node)

        if right_precedence != 0 and (
            (op_precedence > right_precedence)
            or (op_precedence == right_precedence and self._is_left_associative(op_node))
        ):
            self.code += '('
            self._expression(right_node)
            self.code += ')'
        else:
            self._expression(right_node)

    def token_break(self):
        if len(self.code) == 0:
            return

        if self.code[-1] not in '[]{}() :"\'=\n\t<>|^&+-*@/%;,':
            self.code += ' '

    def visit_JoinedStr(self, node):
        assert isinstance(node, ast.JoinedStr)

        import python_minifier.f_string

        self.token_break()
        self.code += str(python_minifier.f_string.OuterFString(node))

    def visit_NamedExpr(self, node):
        self._expression(node.target)
        self.code += ':='
        self._expression(node.value)

    def visit_Await(self, node):
        assert isinstance(node, ast.Await)
        self.token_break()
        self.code += 'await'
        self._rhs(node.value, node)



import ast

from python_minifier.transforms.suite_transformer import SuiteTransformer


class RemovePass(SuiteTransformer):
    """
    Remove Pass keywords from source

    If a statement is syntactically necessary, use an empty expression instead
    """

    def __call__(self, node):
        return self.visit(node)

    def suite(self, node_list, parent):
        without_pass = [self.visit(a) for a in filter(lambda n: not self.is_node(n, ast.Pass), node_list)]

        if len(without_pass) == 0:
            if isinstance(parent, ast.Module):
                return []
            else:
                return [self.add_child(ast.Expr(value=ast.Num(0)), parent=parent)]

        return without_pass



import ast
import sys

import pytest

from python_minifier import unparse
from python_minifier.ast_compare import compare_ast
from python_minifier.rename import add_namespace, bind_names, resolve_names, rename, rename_literals, allow_rename_locals, allow_rename_globals

def hoist(source):
    module = ast.parse(source)
    add_namespace(module)
    bind_names(module)
    resolve_names(module)
    allow_rename_locals(module, False)
    allow_rename_globals(module, False)
    rename_literals(module)
    rename(module)
    print(unparse(module))
    return module

def test_nohoist_single_usage():
    source = '''
a = 'Hello'
'''

    expected = '''
a = 'Hello'
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_multiple_usage():
    source = '''
A = 'Hello'
B = 'Hello'
'''

    expected = '''
C = 'Hello'
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_no_hoist_multiple_small():
    source = '''
A = ''
B = ''
'''

    expected = '''
A = ''
B = ''
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_multiple_small():
    source = '''
A = '.'
B = '.'
C = '.'
D = '.'
E = '.'
F = '.'
G = '.'
'''

    expected = '''
H = '.'
A = H
B = H
C = H
D = H
E = H
F = H
G = H
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_insert_after_docstring():
    source = '''
"Hello this is a docstring"
import collections
A = 'Hello'
B = 'Hello'
'''

    expected = '''
"Hello this is a docstring"
C = 'Hello'
import collections
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)


def test_hoist_insert_after_future():
    source = '''
from  __future__ import print_function
import collections
A = 'Hello'
B = 'Hello'
'''

    expected = '''
from __future__ import print_function
C = 'Hello'
import collections
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_insert_after_future_docstring():
    source = '''
"Hello this is a docstring"
from  __future__ import print_function
import collections
A = 'Hello'
B = 'Hello'
'''

    expected = '''
"Hello this is a docstring"
from __future__ import print_function
C = 'Hello'
import collections
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_bytes():
    source = '''
"Hello this is a docstring"
from  __future__ import print_function
import collections
A = b'Hello'
B = b'Hello'
'''

    expected = '''
"Hello this is a docstring"
from __future__ import print_function
C = b'Hello'
import collections
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_after_multiple_future():
    source = '''
"Hello this is a docstring"
from __future__ import print_function
from __future__ import sausages
import collections
A = b'Hello'
B = b'Hello'
'''

    expected = '''
"Hello this is a docstring"
from __future__ import print_function
from __future__ import sausages
C = b'Hello'
import collections
A = C
B = C
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_class():
    source = '''
class a:
    b = 'Hello'
    c = 'Hello'
'''

    expected = '''
A = 'Hello'
class a:
    b = A
    c = A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_class_to_func():
    source = '''
def z():
    class a:
        b = 'Hello'
        c = 'Hello'
'''

    expected = '''
def z():
    A = 'Hello'
    class a:
        b = A
        c = A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_func_to_func():
    source = '''
def z():
    def y():
        return 'Hello'
    def x():
        return 'Hello'
'''

    expected = '''
def z():
    A = 'Hello'
    def y():
        return A
    def x():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_in_func():
    source = '''
def z():
    def y():
        return 'Hello'
        return 'Hello'
'''

    expected = '''
def z():
    def y():
        A = 'Hello'
        return A
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_over_class():
    source = '''
class a:
    def a():
        return 'Hello'
    def c():
        return 'Hello'
'''

    expected = '''
A = 'Hello'
class a:
    def a():
        return A
    def c():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_generator():
    source = '''
class a:
    def a():
        for i in (a for a in 'Hello'):
            pass
        for i in (a for a in 'World'):
            pass            
        return 'World'
    def c():
        return 'Hello'
'''

    expected = '''
A = 'Hello'
class a:
    def a():
        B = 'World'
        for i in (a for a in A):
            pass
        for i in (a for a in B):
            pass            
        return B
    def c():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_listcomp():
    source = '''
class a:
    def a():
        for i in [a for a in 'Hello']:
            pass
        for i in [a for a in 'World']:
            pass            
        return 'World'
    def c():
        return 'Hello'
'''

    expected = '''
A = 'Hello'
class a:
    def a():
        B = 'World'
        for i in [a for a in A]:
            pass
        for i in [a for a in B]:
            pass            
        return B
    def c():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)


def test_hoist_from_dictcomp():
    if sys.version_info < (2, 7):
        pytest.skip('No DictComp in python < 2.7')

    source = '''
class a:
    def a():
        for i in {a: a for a in 'Hello'}:
            pass
        for i in {a: a for a in 'World'}:
            pass            
        return 'World'
    def c():
        return 'Hello'
'''

    expected = '''
A = 'Hello'
class a:
    def a():
        B = 'World'
        for i in {a: a for a in A}:
            pass
        for i in {a: a for a in B}:
            pass            
        return B
    def c():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_setcomp():
    if sys.version_info < (2, 7):
        pytest.skip('No SetComp in python < 2.7')

    source = '''
class a:
    def a():
        for i in {a for a in 'Hello' + 'Hello' + 'World'}:
            pass
        return 'World'    
    def c():
        return 'World'
'''

    expected = '''
A = 'World'
class a:
    def a():
        B = 'Hello'
        for i in {a for a in B + B + A}:
            pass  
        return A
    def c():
        return A
'''

    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoist_from_lambda():
    source = '''
class a:
    def a():
        lambda: 'Hello' + 'Hello' + 'World'
    def c():
        return 'World'
'''

    expected = '''
A = 'World'
class a:
    def a():
        B = 'Hello'
        lambda: B + B + A
    def c():
        return A
'''
    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)

def test_hoisted_types_py3():
    if sys.version_info < (3, 0):
        pytest.skip('Python 3 only')

    source = '''
a = 'Hello' + 'Hello'
b = u'Hello' + u'Hello'
c = b'Hello' + b'Hello'
'''

    expected = '''
B = b'Hello'
A = 'Hello'
a = A + A
b = A + A
c = B + B
'''
    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)


def test_hoisted_types_py2():
    if sys.version_info >= (3, 0):
        pytest.skip('Python 2 only')

    source = '''
a = 'Hello' + 'Hello'
b = u'Hello' + u'Hello'
c = b'Hello' + b'Hello'
'''

    expected = '''
B = u'Hello'
A = 'Hello'
a = A + A
b = B + B
c = A + A
'''
    expected_ast = ast.parse(expected)
    actual_ast = hoist(source)
    compare_ast(expected_ast, actual_ast)



"""
This package transforms python source code strings or ast.Module Nodes into
a 'minified' representation of the same source code.

"""

import ast

from python_minifier.ast_compare import CompareError, compare_ast
from python_minifier.module_printer import ModulePrinter
from python_minifier.rename import (
    rename_literals,
    bind_names,
    resolve_names,
    rename,
    allow_rename_globals,
    allow_rename_locals,
    add_namespace,
)
from python_minifier.transforms.combine_imports import CombineImports
from python_minifier.transforms.remove_annotations import RemoveAnnotations
from python_minifier.transforms.remove_literal_statements import RemoveLiteralStatements
from python_minifier.transforms.remove_object_base import RemoveObject
from python_minifier.transforms.remove_pass import RemovePass
from python_minifier.transforms.remove_posargs import remove_posargs


class UnstableMinification(RuntimeError):
    """
    Raised when a minified module differs from the original module in an unexpected way.

    This is raised when the minifier generates source code that doesn't parse back into the
    original module (after known transformations).
    This should never occur and is a bug.

    """

    def __init__(self, exception, source, minified):
        self.exception = exception
        self.source = source
        self.minified = minified

    def __str__(self):
        return 'Minification was unstable! Please create an issue at https://github.com/dflook/python-minifier/issues'


def minify(
    source,
    filename=None,
    remove_annotations=True,
    remove_pass=True,
    remove_literal_statements=False,
    combine_imports=True,
    hoist_literals=True,
    rename_locals=True,
    preserve_locals=None,
    rename_globals=False,
    preserve_globals=None,
    remove_object_base=True,
    convert_posargs_to_args=True,
):
    """
    Minify a python module

    The module is transformed according the the arguments.
    If all transformation arguments are False, no transformations are made to the AST, the returned string will
    parse into exactly the same module.

    Using the default arguments only transformations that are always or almost always safe are enabled.

    :param str source: The python module source code
    :param str filename: The original source filename if known

    :param bool remove_annotations: If type annotations should be removed where possible
    :param bool remove_pass: If Pass statements should be removed where possible
    :param bool remove_literal_statements: If statements consisting of a single literal should be removed, including docstrings
    :param bool combine_imports: Combine adjacent import statements where possible
    :param bool hoist_literals: If str and byte literals may be hoisted to the module level where possible.
    :param bool rename_locals: If local names may be shortened
    :param preserve_locals: Locals names to leave unchanged when rename_locals is True
    :type preserve_locals: list[str]
    :param bool rename_globals: If global names may be shortened
    :param preserve_globals: Global names to leave unchanged when rename_globals is True
    :type preserve_globals: list[str]
    :param bool remove_object_base: If object as a base class may be removed
    :param bool convert_posargs_to_args: If positional-only arguments will be converted to normal arguments

    :rtype: str

    """

    filename = filename or 'python_minifier.minify source'

    # This will raise if the source file can't be parsed
    module = ast.parse(source, filename)

    add_namespace(module)

    if remove_literal_statements:
        module = RemoveLiteralStatements()(module)

    if combine_imports:
        module = CombineImports()(module)

    if remove_annotations:
        module = RemoveAnnotations()(module)

    if remove_pass:
        module = RemovePass()(module)

    if remove_object_base:
        module = RemoveObject()(module)

    bind_names(module)
    resolve_names(module)

    if module.tainted:
        rename_globals = False
        rename_locals = False

    allow_rename_locals(module, rename_locals, preserve_locals)
    allow_rename_globals(module, rename_globals, preserve_globals)

    if hoist_literals:
        rename_literals(module)

    rename(module, prefix_globals=not rename_globals, preserved_globals=preserve_globals)

    if convert_posargs_to_args:
        module = remove_posargs(module)

    return unparse(module)


def unparse(module):
    """
    Turn a module AST into python code

    This returns an exact representation of the given module,
    such that it can be parsed back into the same AST.

    :param module: The module to turn into python code
    :type: module: :class:`ast.Module`
    :rtype: str

    """

    assert isinstance(module, ast.Module)

    printer = ModulePrinter()
    printer(module)

    try:
        minified_module = ast.parse(printer.code, 'python_minifier.unparse output')
    except SyntaxError as syntax_error:
        raise UnstableMinification(syntax_error, '', printer.code)

    try:
        compare_ast(module, minified_module)
    except CompareError as compare_error:
        raise UnstableMinification(compare_error, '', printer.code)

    return printer.code


def awslambda(source, filename=None, entrypoint=None):
    """
    Minify a python module for use as an AWS Lambda function

    This returns a string suitable for embedding in a cloudformation template.
    When minifying, all transformations are enabled.

    :param str source: The python module source code
    :param str filename: The original source filename if known
    :param entrypoint: The lambda entrypoint function
    :type entrypoint: str or NoneType
    :rtype: str

    """

    rename_globals = True
    if entrypoint is None:
        rename_globals = False

    return minify(
        source, filename, remove_literal_statements=True, rename_globals=rename_globals, preserve_globals=[entrypoint],
    )



import ast

from python_minifier.rename.binding import Binding
from python_minifier.rename.util import insert, is_str
from python_minifier.transforms.suite_transformer import NodeVisitor


def replace(old_node, new_node):
    parent = old_node.parent
    new_node.parent = parent
    new_node.namespace = old_node.namespace

    for field, old_value in ast.iter_fields(parent):
        if old_value is old_node:
            setattr(parent, field, new_node)
            return

        if isinstance(old_value, list):
            for i, value in enumerate(old_value):
                if value is old_node:
                    old_value[i] = new_node
                    return


class HoistedBinding(Binding):
    def __init__(self, value_node, *args, **kwargs):
        super(HoistedBinding, self).__init__(*args, **kwargs)
        self._value_node = value_node
        self._local_namespace = None

    def __eq__(self, other):
        return type(self.value) is type(other.value) and self.value == other.value

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(repr(self.value))

    def set_local_namespace(self, node):
        self._local_namespace = node

    @property
    def value(self):
        if is_str(self._value_node) or (hasattr(ast, 'Bytes') and isinstance(self._value_node, ast.Bytes)):
            return self._value_node.s
        else:
            return self._value_node.value

    def __repr__(self):
        return self.__class__.__name__ + '(value=%r)' % self.value

    def rename(self, new_name):

        for node in self.references:
            replace(node, ast.Name(id=new_name, ctx=ast.Load()))

        self._local_namespace.body = list(
            insert(
                self._local_namespace.body,
                ast.Assign(targets=[ast.Name(id=new_name, ctx=ast.Store())], value=self._value_node),
            )
        )

        self._name = new_name

    def should_rename(self, new_name):
        current_cost = len(self.references) * len(repr(self.value))
        rename_cost = (len(self.references) * len(new_name)) + self._rename_cost

        return rename_cost <= current_cost


class HoistedValue(object):
    """
    HoistedValue comparator object

    This is for wrapping a value in a set or dict key, and
    ensures different types hash differently, even if they compare equal.

    The problematic values are str/bytes/unicode and int/float.

    """

    def __init__(self, value):
        self._value = value

    def __hash__(self):
        return hash(str(type(self._value)) + str(hash(self._value)))

    def __eq__(self, other):
        return type(self._value) == type(other._value) and self._value == other._value

    def __ne__(self, other):
        return not self == other


class HoistLiterals(NodeVisitor):
    """
    Hoist literal strings to module level variables
    """

    def __call__(self, module):
        self.module = module
        self._hoisted = {}
        self.visit(module)
        self.place_bindings()

    def nearest_function_namespace(self, node):
        """
        Return the namespace node for the nearest function scope.

        This could be itself.

        :param node: The node to get the function namespace of
        :type node: ast.Node
        :rtype: ast.Node

        """

        if isinstance(node.namespace, (ast.FunctionDef, ast.Module)):
            return node.namespace
        if hasattr(ast, 'AsyncFunctionDef') and isinstance(node.namespace, ast.AsyncFunctionDef):
            return node.namespace
        return self.nearest_function_namespace(node.namespace)

    def namespace_path(self, node):
        """
        Return the path of function namespace nodes from the module node down to the input node

        With the source module:
        >>> def a():
        ...   def b():
        ...     c

        >>> namespace_path(c)
        [a, b, c]

        :param node:
        :type node: ast.Node
        :rtype: list[ast.AST]

        """

        l = []

        while True:
            namespace = self.nearest_function_namespace(node)
            l.insert(0, namespace)

            if isinstance(namespace, ast.Module):
                break

            node = namespace

        return l

    def common_path(self, n1_path, n2_path):

        path = []
        for n1_step, n2_step in zip(n1_path, n2_path):
            if n1_step is not n2_step:
                return path
            path.append(n1_step)
        return path

    def place_bindings(self):
        for binding in self._hoisted.values():

            namespace_path = []

            for node in binding.references:
                if not namespace_path:
                    namespace_path = self.namespace_path(node)
                else:
                    namespace_path = self.common_path(namespace_path, self.namespace_path(node))

            namespace_path[-1].bindings.append(binding)
            binding.set_local_namespace(namespace_path[-1])

    def get_binding(self, value, node):
        hoisted_value = HoistedValue(value)
        if hoisted_value in self._hoisted:
            return self._hoisted[hoisted_value]

        binding = HoistedBinding(node, rename_cost=len('=' + repr(value)))
        self._hoisted[hoisted_value] = binding
        return binding

    def visit_Str(self, node):

        if isinstance(node.parent, ast.Expr):
            # This is literal statement
            # The RemoveLiteralStatements transformer must have left it here, so ignore it.
            return

        self.get_binding(node.s, node).add_reference(node)

    def visit_Bytes(self, node):
        self.visit_Str(node)

    def visit_JoinedStr(self, node):
        for v in node.values:
            if is_str(v):
                # Can't hoist this!
                continue
            else:
                self.visit(v)

    def visit_NameConstant(self, node):
        self.get_binding(node.value, node).add_reference(node)


def rename_literals(module):
    HoistLiterals()(module)



from python_minifier.rename.bind_names import bind_names
from python_minifier.rename.mapper import add_namespace
from python_minifier.rename.rename_literals import rename_literals
from python_minifier.rename.renamer import rename
from python_minifier.rename.resolve_names import resolve_names
from python_minifier.rename.util import allow_rename_locals, allow_rename_globals



import ast

from python_minifier.transforms.suite_transformer import SuiteTransformer


def find_doc(node):

    if isinstance(node, ast.Attribute):
        if node.attr == '__doc__':
            raise ValueError('__doc__ found!')

    for child in ast.iter_child_nodes(node):
        find_doc(child)


def _doc_in_module(module):
    try:
        find_doc(module)
        return False
    except:
        return True


class RemoveLiteralStatements(SuiteTransformer):
    """
    Remove literal expressions from the code

    This includes docstrings
    """

    def __call__(self, node):
        if _doc_in_module(node):
            return node
        return self.visit(node)

    def visit_Module(self, node):
        for binding in node.bindings:
            if binding.name == '__doc__':
                node.body = [self.visit(a) for a in node.body]
                return node

        node.body = self.suite(node.body, parent=node)
        return node

    def is_literal_statement(self, node):
        if not isinstance(node, ast.Expr):
            return False

        return self.is_node(node.value, (ast.Num, ast.Str, 'NameConstant', 'Bytes'))

    def suite(self, node_list, parent):
        without_literals = [self.visit(n) for n in node_list if not self.is_literal_statement(n)]

        if len(without_literals) == 0:
            if isinstance(parent, ast.Module):
                return []
            else:
                return [self.add_child(ast.Expr(value=ast.Num(0)), parent=parent)]

        return without_literals



import ast
import sys

from python_minifier.transforms.suite_transformer import SuiteTransformer


class RemoveAnnotations(SuiteTransformer):
    """
    Remove type annotations from source
    """

    def __call__(self, node):
        if sys.version_info < (3, 0):
            return node
        return self.visit(node)

    def visit_FunctionDef(self, node):
        node.args = self.visit_arguments(node.args)
        node.body = self.suite(node.body, parent=node)
        node.decorator_list = [self.visit(d) for d in node.decorator_list]

        if hasattr(node, 'returns'):
            node.returns = None

        return node

    def visit_arguments(self, node):
        assert isinstance(node, ast.arguments)

        if hasattr(node, 'posonlyargs') and node.posonlyargs:
            node.posonlyargs = [self.visit_arg(a) for a in node.posonlyargs]

        if node.args:
            node.args = [self.visit_arg(a) for a in node.args]

        if hasattr(node, 'kwonlyargs') and node.kwonlyargs:
            node.kwonlyargs = [self.visit_arg(a) for a in node.kwonlyargs]

        if hasattr(node, 'varargannotation'):
            node.varargannotation = None
        else:
            if node.vararg:
                node.vararg = self.visit_arg(node.vararg)

        if hasattr(node, 'kwargannotation'):
            node.kwargannotation = None
        else:
            if node.kwarg:
                node.kwarg = self.visit_arg(node.kwarg)

        return node

    def visit_arg(self, node):
        node.annotation = None
        return node

    def visit_AnnAssign(self, node):
        def is_dataclass_field(node):
            if sys.version_info < (3, 7):
                return False

            if not isinstance(node.parent, ast.ClassDef):
                return False

            if len(node.parent.decorator_list) == 0:
                return False

            for node in node.parent.decorator_list:
                if isinstance(node, ast.Name) and node.id == 'dataclass':
                    return True
                elif isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id == 'dataclass':
                    return True

            return False

        if is_dataclass_field(node):
            return node
        elif node.value:
            return self.add_child(ast.Assign([node.target], node.value), parent=node.parent)
        else:
            # Valueless annotations cause the interpreter to treat the variable as a local.
            # I don't know of another way to do that without assigning to it, so
            # keep it as an AnnAssign, but replace the annotation with '0'

            node.annotation = self.add_child(ast.Num(0), parent=node.parent)
            return node



import ast

from python_minifier.transforms.suite_transformer import SuiteTransformer


class CombineImports(SuiteTransformer):
    """
    Combine multiple import statements where possible

    This doesn't change the order of imports

    """

    def _combine_import(self, node_list, parent):

        alias = []

        for statement in node_list:
            if isinstance(statement, ast.Import):
                alias += statement.names
            else:
                if alias:
                    yield self.add_child(ast.Import(names=alias), parent=parent)
                    alias = []

                yield statement

        if alias:
            yield self.add_child(ast.Import(names=alias), parent=parent)

    def _combine_import_from(self, node_list, parent):

        prev_import = None
        alias = []

        def combine(statement):
            if not isinstance(statement, ast.ImportFrom):
                return False

            if len(statement.names) == 1 and statement.names[0].name == '*':
                return False

            if prev_import is None:
                return True

            if statement.module == prev_import.module and statement.level == prev_import.level:
                return True

            return False

        for statement in node_list:
            if combine(statement):
                prev_import = statement
                alias += statement.names
            else:
                if alias:
                    yield self.add_child(
                        ast.ImportFrom(module=prev_import.module, names=alias, level=prev_import.level), parent=parent
                    )
                    alias = []

                yield statement

        if alias:
            yield self.add_child(
                ast.ImportFrom(module=prev_import.module, names=alias, level=prev_import.level), parent=parent
            )

    def suite(self, node_list, parent):
        a = list(self._combine_import(node_list, parent))
        b = list(self._combine_import_from(a, parent))

        return [self.visit(n) for n in b]



import ast
import sys
import pytest
from python_minifier import unparse
from python_minifier.ast_compare import compare_ast
from python_minifier.transforms.remove_posargs import remove_posargs

def test_pep():
    if sys.version_info < (3, 8):
        pytest.skip('No Assignment expressions in python < 3.8')

    source = '''
def name(p1, p2, /, p_or_kw, *, kw): pass
def name(p1, p2=None, /, p_or_kw=None, *, kw): pass
def name(p1, p2=None, /, *, kw): pass
def name(p1, p2=None, /): pass
def name(p1, p2, /, p_or_kw): pass
def name(p1, p2, /): pass
def name(p_or_kw, *, kw): pass
def name(*, kw): pass

def standard_arg(arg):
    print(arg)
def pos_only_arg(arg, /):
    print(arg)
def kwd_only_arg(*, arg):
    print(arg)
def combined_example(pos_only, /, standard, *, kwd_only):
    print(pos_only, standard, kwd_only)
'''

    expected_ast = ast.parse(source)
    actual_ast = unparse(expected_ast)
    compare_ast(expected_ast, ast.parse(actual_ast))

def test_convert():
    if sys.version_info < (3, 8):
        pytest.skip('No Assignment expressions in python < 3.8')

    source = '''
def name(p1, p2, /, p_or_kw, *, kw): pass
def name(p1, p2=None, /, p_or_kw=None, *, kw): pass
def name(p1, p2=None, /, *, kw): pass
def name(p1, p2=None, /): pass
def name(p1, p2, /, p_or_kw): pass
def name(p1, p2, /): pass
def name(p_or_kw, *, kw): pass
def name(*, kw): pass

def standard_arg(arg):
    print(arg)
def pos_only_arg(arg, /):
    print(arg)
def kwd_only_arg(*, arg):
    print(arg)
def combined_example(pos_only, /, standard, *, kwd_only):
    print(pos_only, standard, kwd_only)
'''

    expected = '''
def name(p1, p2, p_or_kw, *, kw): pass
def name(p1, p2=None, p_or_kw=None, *, kw): pass
def name(p1, p2=None, *, kw): pass
def name(p1, p2=None): pass
def name(p1, p2, p_or_kw): pass
def name(p1, p2): pass
def name(p_or_kw, *, kw): pass
def name(*, kw): pass

def standard_arg(arg):
    print(arg)
def pos_only_arg(arg):
    print(arg)
def kwd_only_arg(*, arg):
    print(arg)
def combined_example(pos_only, standard, *, kwd_only):
    print(pos_only, standard, kwd_only)
'''

    expected_ast = ast.parse(expected)
    actual_ast = unparse(remove_posargs(ast.parse(source)))
    compare_ast(expected_ast, ast.parse(actual_ast))


"""
Test for renaming of builtins

This assumes the standard NameAssigner and name_generator
"""

import ast
import sys

import pytest

from python_minifier import add_namespace, bind_names, resolve_names, allow_rename_locals, allow_rename_globals, \
    compare_ast, rename, CompareError, unparse


def do_rename(source):
    # This will raise if the source file can't be parsed
    module = ast.parse(source, 'test_rename_bultins')
    add_namespace(module)
    bind_names(module)
    resolve_names(module)

    allow_rename_locals(module, True)
    allow_rename_globals(module, True)

    rename(module)

    return module


def assert_code(expected_ast, actual_ast):
    try:
        compare_ast(expected_ast, actual_ast)
    except CompareError as e:
        print(e)
        print(unparse(actual_ast))
        raise


def test_rename_builtins():
    source = '''
sorted()
sorted()
sorted()
sorted()
sorted()
'''
    expected = '''
A=sorted
A()
A()
A()
A()
A()
'''

    expected_ast = ast.parse(expected)
    actual_ast = do_rename(source)
    assert_code(expected_ast, actual_ast)


def test_no_rename_assigned_builtin():
    source = '''
if random.choice([True, False]):
    sorted=str
sorted()
sorted()
sorted()
sorted()
sorted()
'''
    expected = source

    expected_ast = ast.parse(expected)
    actual_ast = do_rename(source)
    assert_code(expected_ast, actual_ast)

def test_rename_local_builtin():
    source = '''
def t():
    sorted()
    sorted()
    sorted()
    sorted()
    sorted()
'''
    expected = '''
A=sorted
def B():
    A()
    A()
    A()
    A()
    A()
'''

    expected_ast = ast.parse(expected)
    actual_ast = do_rename(source)
    assert_code(expected_ast, actual_ast)

def test_no_rename_local_assigned_builtin():
    source = '''
def a():
    if random.choice([True, False]):
        sorted=str
    sorted()
    sorted()
    sorted()
    sorted()
    sorted()
'''

    expected = '''
def A():
    if random.choice([True, False]):
        sorted=str
    sorted()
    sorted()
    sorted()
    sorted()
    sorted()
'''

    expected_ast = ast.parse(expected)
    actual_ast = do_rename(source)
    assert_code(expected_ast, actual_ast)



BACKSLASH = '\\'


class MiniString(object):
    """
    Create a representation of a string object

    :param str string: The string to minify

    """

    def __init__(self, string, quote="'"):
        self._s = string
        self.safe_mode = False
        self.quote = quote

    def __str__(self):
        """
        The smallest python literal representation of a string

        :rtype: str

        """

        if self._s == '':
            return ''

        if len(self.quote) == 1:
            s = self.to_short()
        else:
            s = self.to_long()

        try:
            eval(self.quote + s + self.quote)
        except UnicodeDecodeError:
            if self._safe_mode:
                raise

            self._safe_mode = True

        assert eval(self.quote + s + self.quote) == self._s

        return s

    def to_short(self):
        s = ''

        escaped = {
            '\n': BACKSLASH + 'n',
            '\\': BACKSLASH + BACKSLASH,
            '\a': BACKSLASH + 'a',
            '\b': BACKSLASH + 'b',
            '\f': BACKSLASH + 'f',
            '\r': BACKSLASH + 'r',
            '\t': BACKSLASH + 't',
            '\v': BACKSLASH + 'v',
            '\0': BACKSLASH + 'x00',
            self.quote: BACKSLASH + self.quote,
        }

        for c in self._s:
            if c in escaped.keys():
                s += escaped[c]
            else:
                if self.safe_mode:
                    unicode_value = ord(c)
                    if unicode_value <= 0x7F:
                        s += c
                    elif unicode_value <= 0xFFFF:
                        s += BACKSLASH + 'u' + format(unicode_value, '04x')
                    else:
                        s += BACKSLASH + 'U' + format(unicode_value, '08x')
                else:
                    s += c

        return s

    def to_long(self):
        s = ''

        escaped = {
            '\\': BACKSLASH + BACKSLASH,
            '\a': BACKSLASH + 'a',
            '\b': BACKSLASH + 'b',
            '\f': BACKSLASH + 'f',
            '\r': BACKSLASH + 'r',
            '\t': BACKSLASH + 't',
            '\v': BACKSLASH + 'v',
            '\0': BACKSLASH + 'x00',
            self.quote[0]: BACKSLASH + self.quote[0],
        }

        for c in self._s:
            if c in escaped.keys():
                s += escaped[c]
            else:
                if self.safe_mode:
                    unicode_value = ord(c)
                    if unicode_value <= 0x7F:
                        s += c
                    elif unicode_value <= 0xFFFF:
                        s += BACKSLASH + 'u' + format(unicode_value, '04x')
                    else:
                        s += BACKSLASH + 'U' + format(unicode_value, '08x')
                else:
                    s += c

        return s


class MiniBytes(object):
    """
    Create a representation of a bytes object

    :param bytes string: The string to minify

    """

    def __init__(self, string, quote="'"):
        self._b = string
        self.quote = quote

    def __str__(self):
        """
        The smallest python literal representation of a string

        :rtype: str

        """

        if self._b == b'':
            return ''

        if len(self.quote) == 1:
            s = self.to_short()
        else:
            s = self.to_long()

        assert eval('b' + self.quote + s + self.quote) == self._b

        return s

    def to_short(self):
        b = ''

        for c in self._b:
            if c == b'\\':
                b += BACKSLASH
            elif c == b'\n':
                b += BACKSLASH + 'n'
            elif c == self.quote:
                b += BACKSLASH + self.quote
            else:
                if c >= 128:
                    b += BACKSLASH + chr(c)
                else:
                    b += chr(c)

        return b

    def to_long(self):
        b = ''

        for c in self._b:
            if c == b'\\':
                b += BACKSLASH
            elif c == self.quote:
                b += BACKSLASH + self.quote
            else:
                if c >= 128:
                    b += BACKSLASH + chr(c)
                else:
                    b += chr(c)

        return b



import ast
import pytest
import sys

from python_minifier.transforms.remove_object_base import RemoveObject
from python_minifier.ast_compare import compare_ast


def test_remove_object_py3():
    if sys.version_info < (3, 0):
        pytest.skip('This test is python3 only')

    source = '''
class Test(object):
    pass
'''
    expected = '''
class Test:
    pass
'''

    expected_ast = ast.parse(expected)
    actual_ast = RemoveObject()(ast.parse(source))
    compare_ast(expected_ast, actual_ast)

    source = '''
class Test(another_base, object, third_base):
    pass
'''
    expected = '''
class Test(another_base, third_base):
    pass
'''

    expected_ast = ast.parse(expected)
    actual_ast = RemoveObject()(ast.parse(source))
    compare_ast(expected_ast, actual_ast)

    expected_ast = ast.parse(expected)
    actual_ast = RemoveObject()(ast.parse(source))
    compare_ast(expected_ast, actual_ast)

    source = '''
class Test(other_base):
    pass
'''
    expected = source

    expected_ast = ast.parse(expected)
    actual_ast = RemoveObject()(ast.parse(source))
    compare_ast(expected_ast, actual_ast)


def test_no_remove_object_py2():
    if sys.version_info >= (3, 0):
        pytest.skip('This test is python2 only')

    source = '''
class Test(object):
    pass
'''
    expected = '''
class Test(object):
    pass
'''

    expected_ast = ast.parse(expected)
    actual_ast = RemoveObject()(ast.parse(source))
    compare_ast(expected_ast, actual_ast)




import ast

from python_minifier import add_namespace, bind_names, resolve_names
from python_minifier.transforms.remove_literal_statements import RemoveLiteralStatements
from python_minifier.ast_compare import compare_ast

def remove_literals(source):
    module = ast.parse(source, 'test_remove_literal_statements')

    add_namespace(module)
    bind_names(module)
    resolve_names(module)
    return RemoveLiteralStatements()(module)

def test_remove_literal_num():
    source = '213'
    expected = ''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_literal_str():
    source = '"hello"'
    expected = ''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)

def test_complex():
    source = '''
"module docstring"
a = 'hello'

def t():
    "function docstring"
    a = 2
    0
    2
    'sadfsaf'
    def g():
        "just a docstring"

'''
    expected = '''
a = 'hello'
def t():
    a=2
    def g():
        0
'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_literals(source)
    compare_ast(expected_ast, actual_ast)



import ast
import sys
import pytest
from python_minifier import unparse
from python_minifier.ast_compare import compare_ast

def test_await_fstring():
    if sys.version_info < (3, 7):
        pytest.skip('Await in f-string expressions not allowed in python < 3.7')

    source = '''
async def a(): return 'hello'
async def b(): return f'{await b()}'
'''

    expected_ast = ast.parse(source)
    actual_ast = unparse(expected_ast)
    compare_ast(expected_ast, ast.parse(actual_ast))



import ast
import sys
import pytest

from python_minifier import add_namespace
from python_minifier.transforms.remove_annotations import RemoveAnnotations
from python_minifier.ast_compare import compare_ast

def remove_annotations(source):
    module = ast.parse(source)
    add_namespace(module)
    RemoveAnnotations()(module)
    return module

def test_AnnAssign():

    if sys.version_info < (3, 6):
        pytest.skip('Variable annotation unsupported in python < 3.6')

    source = '''a :str = 1
b = 2
c : int = 3'''
    expected = '''a = 1
b = 2
c = 3'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)

def test_FunctionDef():
    if sys.version_info < (3,):
        pytest.skip('Annotation unsupported in python < 3.0')

    source = '''def test(a: str, b: int=1, *c: hello, **aws: crap) -> None:
    pass'''
    expected = '''def test(a, b=1, *c, **aws):
    pass'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)

def test_AsyncFunctionDef():
    if sys.version_info < (3, 6):
        pytest.skip('Async function unsupported in python < 3.5')

    source = '''async def test(a: str, b: int=1, *c: hello, **aws: crap) -> None:
    pass'''
    expected = '''async def test(a, b=1, *c, **aws):
    pass'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)

def test_AnnAssign_novalue():
    if sys.version_info < (3, 6):
        pytest.skip('Variable annotation unsupported in python < 3.6')

    source = '''a :str
'''
    expected = '''a:0'''

    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)

def test_no_remove_dataclass():
    if sys.version_info < (3, 6):
        pytest.skip('annotations unavailable in python < 3.6')

    if sys.version_info < (3, 7):
        pytest.skip('dataclass unavailable in python < 3.7')

    source = '''
@dataclass
class MyClass:
    myfield: int
    mysecondfile: str
'''
    expected = source

    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)

def test_remove_dataclass():
    if sys.version_info < (3, 6):
        pytest.skip('annotations unavailable in python < 3.6')

    if sys.version_info >= (3, 7):
        pytest.skip('dataclass available in python >= 3.7')

    source = '''
@dataclass
class MyClass:
    myfield: int
    mysecondfile: str
'''
    expected = '''
@dataclass
class MyClass:
    myfield: 0
    mysecondfile: 0
'''
    expected_ast = ast.parse(expected)
    actual_ast = remove_annotations(source)
    compare_ast(expected_ast, actual_ast)



from __future__ import print_function

import sys

import argparse
from pkg_resources import get_distribution, DistributionNotFound

from python_minifier import minify

try:
    version = get_distribution('python_minifier').version
except DistributionNotFound:
    version = '0.0.0'


def main():

    parser = argparse.ArgumentParser(prog='pyminify', description='Minify Python source')

    parser.add_argument('path', type=str, help='The source file to minify. Use "-" to read from stdin')

    parser.add_argument(
        '--no-combine-imports',
        action='store_false',
        help='Disable combining adjacent import statements',
        dest='combine_imports',
    )
    parser.add_argument(
        '--no-remove-pass',
        action='store_false',
        default=True,
        help='Disable removing Pass statements',
        dest='remove_pass',
    )
    parser.add_argument(
        '--remove-literal-statements',
        action='store_true',
        help='Enable removing statements that are just a literal (including docstrings)',
        dest='remove_literal_statements',
    )
    parser.add_argument(
        '--no-remove-annotations',
        action='store_false',
        help='Disable removing function and variable annotations',
        dest='remove_annotations',
    )
    parser.add_argument(
        '--no-hoist-literals',
        action='store_false',
        help='Disable replacing string and bytes literals with variables',
        dest='hoist_literals',
    )
    parser.add_argument(
        '--no-rename-locals', action='store_false', help='Disable shortening of local names', dest='rename_locals'
    )
    parser.add_argument(
        '--preserve-locals',
        type=str,
        action='append',
        help='Comma separated list of local names that will not be shortened',
        dest='preserve_locals',
    )
    parser.add_argument(
        '--rename-globals', action='store_true', help='Enable shortening of global names', dest='rename_globals'
    )
    parser.add_argument(
        '--preserve-globals',
        type=str,
        action='append',
        help='Comma separated list of global names that will not be shortened',
        dest='preserve_globals',
    )
    parser.add_argument(
        '--no-remove-object-base',
        action='store_false',
        help='Disable removing object from base class list',
        dest='remove_object_base',
    )
    parser.add_argument(
        '--no-convert-posargs-to-args',
        action='store_false',
        help='Disable converting positional only arguments to normal arguments',
        dest='convert_posargs_to_args',
    )

    parser.add_argument('-v', '--version', action='version', version=version)

    args = parser.parse_args()

    if args.path == '-':
        source = sys.stdin.read()
    else:
        with open(args.path, 'rb') as f:
            source = f.read()

    preserve_globals = []
    if args.preserve_globals:
        for arg in args.preserve_globals:
            names = [name.strip() for name in arg.split(',') if name]
            preserve_globals.extend(names)

    preserve_locals = []
    if args.preserve_locals:
        for arg in args.preserve_locals:
            names = [name.strip() for name in arg.split(',') if name]
            preserve_locals.extend(names)

    sys.stdout.write(
        minify(
            source,
            filename=args.path,
            combine_imports=args.combine_imports,
            remove_pass=args.remove_pass,
            remove_annotations=args.remove_annotations,
            remove_literal_statements=args.remove_literal_statements,
            hoist_literals=args.hoist_literals,
            rename_locals=args.rename_locals,
            preserve_locals=preserve_locals,
            rename_globals=args.rename_globals,
            preserve_globals=preserve_globals,
            remove_object_base=args.remove_object_base,
            convert_posargs_to_args=args.convert_posargs_to_args,
        )
    )


if __name__ == '__main__':
    main()



import sys
import os.path
from setuptools import setup, find_packages

readme_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md')
with open(readme_path) as f:
    long_desc = f.read()

install_requires = []
if sys.version_info < (2, 7):
    install_requires = ['argparse']

setup(
    name='python_minifier',
    description='Transform Python source code into it\'s most compact representation',
    author='Daniel Flook',
    author_email='daniel@flook.org',
    url='https://github.com/dflook/python-minifier',
    license='MIT',
    project_urls={
        'Issues': 'https://github.com/dflook/python-minifier/issues',
        'Documentation': 'https://dflook.github.io/python-minifier/',
    },
    keywords='minify minifier',

    
    package_dir={'': 'src'},
    packages=find_packages('src'),
    long_description=long_desc,
    long_description_content_type='text/markdown',

    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <3.10',
    version='2.4.0',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Developers',
        'Topic :: Software Development'
    ],

    entry_points = {
        'console_scripts': ['pyminify=python_minifier.__main__:main']
    },

    install_requires=install_requires,

    zip_safe=True
)



import itertools
import keyword
import random
import string

from python_minifier.rename.util import builtins


def random_generator(length=40):
    valid_first = string.ascii_uppercase + string.ascii_lowercase
    valid_rest = string.digits + valid_first + '_'

    while True:
        first = [random.choice(valid_first)]
        rest = [random.choice(valid_rest) for i in range(length - 1)]
        yield ''.join(first + rest)


def name_generator():
    valid_first = string.ascii_uppercase + string.ascii_lowercase
    valid_rest = string.digits + valid_first + '_'

    for c in valid_first:
        yield c

    for length in itertools.count(1):
        for first in valid_first:
            for rest in itertools.product(valid_rest, repeat=length):
                name = first
                name += ''.join(rest)
                yield name


def name_filter():
    """
    Yield all valid python identifiers

    Name are returned sorted by length, then string sort order.

    Names that already have meaning in python (keywords and builtins)
    will not be included in the output.

    :rtype: Iterable[str]

    """

    reserved = keyword.kwlist + dir(builtins)

    for name in name_generator():
        if name not in reserved:
            yield name



import ast


def remove_posargs(node):
    if isinstance(node, ast.arguments):
        if hasattr(node, 'posonlyargs'):
            node.args = node.posonlyargs + node.args
            node.posonlyargs = []

    for child in ast.iter_child_nodes(node):
        remove_posargs(child)

    return node



"""
For each node in an AST set the namespace to use for name binding and resolution
"""

import ast

from python_minifier.rename.util import is_namespace


def add_parent_to_arguments(arguments, func):
    arguments.parent = func
    arguments.namespace = func

    for arg in getattr(arguments, 'posonlyargs', []) + arguments.args:
        add_parent(arg, arguments, func)
        if hasattr(arg, 'annotation') and arg.annotation is not None:
            add_parent(arg.annotation, arguments, func.namespace)

    if hasattr(arguments, 'kwonlyargs'):
        for arg in arguments.kwonlyargs:
            add_parent(arg, arguments, func)
            if arg.annotation is not None:
                add_parent(arg.annotation, arguments, func.namespace)

        for node in arguments.kw_defaults:
            if node is not None:
                add_parent(node, arguments, func.namespace)

    for node in arguments.defaults:
        add_parent(node, arguments, func.namespace)

    if arguments.vararg:
        if hasattr(arguments, 'varargannotation') and arguments.varargannotation is not None:
            add_parent(arguments.varargannotation, arguments, func.namespace)
        elif isinstance(arguments.vararg, str):
            pass
        else:
            add_parent(arguments.vararg, arguments, func)

    if arguments.kwarg:
        if hasattr(arguments, 'kwargannotation') and arguments.kwargannotation is not None:
            add_parent(arguments.kwargannotation, arguments, func.namespace)
        elif isinstance(arguments.kwarg, str):
            pass
        else:
            add_parent(arguments.kwarg, arguments, func)


def add_parent_to_functiondef(functiondef):
    """
    Add correct parent and namespace attributes to functiondef nodes
    """

    if functiondef.args is not None:
        add_parent_to_arguments(functiondef.args, func=functiondef)

    for node in functiondef.body:
        add_parent(node, parent=functiondef, namespace=functiondef)

    for node in functiondef.decorator_list:
        add_parent(node, parent=functiondef, namespace=functiondef.namespace)

    if hasattr(functiondef, 'returns') and functiondef.returns is not None:
        add_parent(functiondef.returns, parent=functiondef, namespace=functiondef.namespace)


def add_parent_to_classdef(classdef):
    """
    Add correct parent and namespace attributes to classdef nodes
    """

    for node in classdef.bases:
        add_parent(node, parent=classdef, namespace=classdef.namespace)

    if hasattr(classdef, 'keywords'):
        for node in classdef.keywords:
            add_parent(node, parent=classdef, namespace=classdef.namespace)

    if hasattr(classdef, 'starargs') and classdef.starargs is not None:
        add_parent(classdef.starargs, parent=classdef, namespace=classdef.namespace)

    if hasattr(classdef, 'kwargs') and classdef.kwargs is not None:
        add_parent(classdef.kwargs, parent=classdef, namespace=classdef.namespace)

    for node in classdef.body:
        add_parent(node, parent=classdef, namespace=classdef)

    for node in classdef.decorator_list:
        add_parent(node, parent=classdef, namespace=classdef.namespace)


def add_parent(node, parent=None, namespace=None):
    """
    Add a parent attribute to child nodes
    Add a namespace attribute to child nodes

    :param node: The tree to add parent and namespace properties to
    :type node: :class:`ast.AST`
    :param parent: The parent node of this node
    :type parent: :class:`ast.AST`
    :param namespace: The namespace Node that this node is in
    :type namespace: ast.Lambda or ast.Module or ast.FunctionDef or ast.AsyncFunctionDef or ast.ClassDef or ast.DictComp or ast.SetComp or ast.ListComp or ast.Generator

    """

    node.parent = parent if parent is not None else node
    node.namespace = namespace if namespace is not None else node

    if is_namespace(node):
        node.bindings = []
        node.global_names = set()
        node.nonlocal_names = set()

        if isinstance(node, ast.FunctionDef) or (
            hasattr(ast, 'AsyncFunctionDef') and isinstance(node, ast.AsyncFunctionDef)
        ):
            add_parent_to_functiondef(node)
        elif isinstance(node, ast.Lambda):
            add_parent_to_arguments(node.args, func=node)
            add_parent(node.body, parent=node, namespace=node)
        elif isinstance(node, ast.ClassDef):
            add_parent_to_classdef(node)
        else:
            for child in ast.iter_child_nodes(node):
                add_parent(child, parent=node, namespace=node)

        return

    if isinstance(node, ast.comprehension):
        add_parent(node.target, parent=node, namespace=namespace)
        add_parent(node.iter, parent=node, namespace=namespace)
        for if_ in node.ifs:
            add_parent(if_, parent=node, namespace=namespace)
        return

    if isinstance(node, ast.Global):
        namespace.global_names.update(node.names)
    if hasattr(ast, 'Nonlocal') and isinstance(node, ast.Nonlocal):
        namespace.nonlocal_names.update(node.names)

    for child in ast.iter_child_nodes(node):
        add_parent(child, parent=node, namespace=namespace)


def add_namespace(module):
    add_parent(module)
